from ..CST import Result1DComplex
from ..CST import ResultTree
from ..CST import ResultTreeItem
import numpy as np
import re

class SParameters:
    ITEM_ROOT = '1D Results\\S-Parameters'
    ITEM_REGEX = '^1D Results\\\\S-Parameters\\\\S(\d+),(\d+)$'
    ITEM_TEMPL = '1D Results\\S-Parameters\\S{},{}'

    def __init__(self, result_tree: ResultTree) -> None:
        self.result_tree = result_tree
        self.project = result_tree.project
        self.com_object = result_tree.com_object

    def are_available(self):
        return self.result_tree.has_item(SParameters.ITEM_ROOT) and self.__has_at_least_one_result()

    def get_available_indices(self):
        res = map(SParameters.get_indices_from_item_name, self.__get_available_items())
        res = SParameters.__remove_none_values(res)
        return res

    def get_by_indices(self, indices):
        if indices in self.get_available_indices():
            return SParameter(self, indices)
        return None

    def get_all(self):
        map_fn = lambda indices: SParameter(self, indices)
        return list(map(map_fn, self.get_available_indices()))

    def get_matrix_size(self):
        return self.get_available_indices()[-1]

    def get_matrix_at_index(self, index: int, result_id = None):
        s_matrix = np.zeros(shape=self.get_matrix_size(), dtype=np.complex128)
        for i in range(s_matrix.shape[0]):
            for j in range(s_matrix.shape[1]):
                s_matrix[i, j] = SParameter(self, (i+1, j+1)).get_value_at_index(index, result_id)
        return s_matrix

    def get_matrix_at_frequency(self, frequency: float, result_id = None):
        index = SParameter(self, (1, 1)).get_nearest_frequency_index(frequency)
        return self.get_matrix_at_index(index, result_id)

    def get_frequencies(self, result_id = None):
        return SParameter(self, (1, 1)).get_frequencies(result_id)

    def get_num_frequencies(self, result_id = None):
        return SParameter(self, (1, 1)).get_num_frequencies(result_id)

    def get_frequency_at_index(self, index: int, result_id = None):
        return SParameter(self, (1, 1)).get_frequency_at_index(index, result_id)

    def get_nearest_frequency_index(self, frequency: float, result_id = None):
        return SParameter(self, (1, 1)).get_nearest_frequency_index(frequency, result_id)

    def get_matrices(self, result_id = None):
        matrix_size = self.get_matrix_size()
        num_points = self.get_num_frequencies(result_id)
        s_matrix = np.zeros(shape=(num_points, matrix_size[0], matrix_size[1]), dtype=np.complex128)
        for i in range(s_matrix.shape[1]):
            for j in range(s_matrix.shape[2]):
                s_matrix[:, i, j] = SParameter(self, (i+1, j+1)).get_values(result_id)
        return s_matrix

    def __has_at_least_one_result(self):
        return len(self.result_tree.get_first_child_name(SParameters.ITEM_ROOT)) > 0

    def __get_available_items(self):
        res = []
        child_name = self.result_tree.get_first_child_name(SParameters.ITEM_ROOT)
        while len(child_name) > 0:
            res.append(child_name)
            child_name = self.result_tree.get_next_item_name(child_name)
        return res

    def __remove_none_values(arr):
        return list(filter(lambda x: x is not None, arr))

    def get_indices_from_item_name(name):
        x = re.search(SParameters.ITEM_REGEX, name)
        return None if x is None else (int(x.group(1)), int(x.group(2)))

    def get_item_name_from_indices(indices):
        return SParameters.ITEM_TEMPL.format(int(indices[0]), int(indices[1]))

    def get_z_matrix(s_matrix, z_ref: complex = 50):
        E = np.eye(s_matrix.shape[0])
        sqrt_z = E*np.sqrt(z_ref)
        return sqrt_z@np.linalg.solve(E-s_matrix, E+s_matrix)@sqrt_z

    def get_z_matrices(s_matrices, z_ref: complex = 50):
        return np.array(list(map(lambda x: SParameters.get_z_matrix(x, z_ref), s_matrices)))

class SParameter:
    def __init__(self, s_parameters: SParameters, indices: tuple[int, int]) -> None:
        self.s_parameters = s_parameters
        self.indices = indices

    def get_result_ids(self):
        item_name = SParameters.get_item_name_from_indices(self.indices)
        return ResultTreeItem(self.s_parameters.result_tree).get_result_ids(item_name)

    def get_result(self, result_id = None):
        if result_id is None:
            result_id = self.get_result_ids()[0]
        obj = self.__get_result_obj(result_id)
        return Result1DComplex(obj)

    def get_num_frequencies(self, result_id = None) -> int:
        return self.get_result(result_id).get_num_elements()

    def get_frequencies(self, result_id = None):
        if result_id is None:
            result_id = self.get_result_ids()[0]
        return self.get_result(result_id).get_x_values()

    def get_frequency_at_index(self, index: int, result_id = None):
        return self.get_result(result_id).get_x_at(index)

    def get_nearest_frequency_index(self, frequency: float, result_id = None):
        freqs = self.get_frequencies(result_id)
        if frequency < freqs[0] or frequency > freqs[-1]:
            return None
        return np.abs(freqs - frequency).argmin()

    def get_values(self, result_id = None):
        if result_id is None:
            result_id = self.get_result_ids()[0]
        return self.get_result(result_id).get_y_values()

    def get_value_at_index(self, index: int, result_id = None):
        return self.get_result(result_id).get_y_at(index)

    def get_value_at_frequency(self, frequency: float, result_id = None):
        index = self.get_nearest_frequency_index(frequency)
        return self.get_value_at_index(index, result_id)

    def __get_result_obj(self, result_id):
        item_name = SParameters.get_item_name_from_indices(self.indices)
        return ResultTreeItem(self.s_parameters.result_tree).get_result(item_name, result_id)


from . import Project
from . import ComObjectWrapper

class Pick(ComObjectWrapper):
    PICK_TYPE_END_POINT = 'EndPoint'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Pick

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def add_edge(self, u1: int, v1: int, w1: int, u2: int, v2: int, w2: int):
        self.invoke_method('AddEdge', u1, v1, w1, u2, v2, w2)

    def delete_edge(self, index: int):
        self.invoke_method('DeleteEdge', index)

    def add_mean_edge(self, indices: list[int]):
        self.invoke_method('MeanEdge', ','.join(str(s) for s in indices))

    def move_edge(self, index: int, dx: float, dy: float, dz: float, keep_original: bool = False):
        self.invoke_method('MoveEdge', index, dx, dy, dz, keep_original)

    def move_edge_in_plane(self, index: int, offset: float, keep_original: bool = False):
        self.invoke_method('MoveEdgeInPlane', index, offset, keep_original)

    def pick_edge_from_picked_points(self, index1: int, index2: int):
        self.invoke_method('PickEdgeFromPickedPoints', index1, index2)

    def clear_all_picks(self):
        self.invoke_method('ClearAllPicks')

    def write_next_pick_to_database(self, id: int):
        self.invoke_method('NextPickToDataBase', id)

    def snap_last_point_to_drawplane(self):
        self.invoke_method('SnapLastPointToDrawplane')

    def pick_point_from_id_on(self, name: str, pick_type: str, id: int):
        self.invoke_method('PickPointFromIdOn', name, pick_type, id)

    def pick_point_from_point_on(self, name: str, pick_type: str, x: float, y: float, z: float):
        self.invoke_method('PickPointFromPointOn', name, pick_type, x, y, z)

    def pick_edge_from_id_on(self, name: str, edge_id: int, vertex_id: int):
        self.invoke_method('PickEdgeFromIdOn', name, edge_id, vertex_id)

    def pick_edge_from_point_on(self, name: str, x: float, y: float, z: float):
        self.invoke_method('PickEdgeFromPointOn', name, x, y, z)

    def pick_face_from_id_on(self, name: str, face_id: int):
        self.invoke_method('PickFaceFromIdOn', name, face_id)

    def pick_face_from_point_on(self, name: str, x: float, y: float, z: float):
        self.invoke_method('PickFaceFromPointOn', name, x, y, z)

    def pick_dangling_edge_chain_from_id(self, shape_name: str, id: int):
        self.invoke_method('PickDanglingEdgeChainFromId', shape_name, id)

    def pick_edge_from_id(self, shape_name: str, edge_id: int, vertex_id: int):
        self.invoke_method('PickEdgeFromId', shape_name, edge_id, vertex_id)

    def pick_edge_from_point(self, shape_name: str, x: float, y: float, z: float):
        self.invoke_method('PickEdgeFromPoint', shape_name, x, y, z)

    def pick_solid_edge_chain_from_id(self, shape_name: str, edge_id: int, face_id: int):
        self.invoke_method('PickSolidEdgeChainFromId', shape_name, edge_id, face_id)

    def pick_face_chain_from_id(self, shape_name: str, face_id: int):
        self.invoke_method('PickFaceChainFromId', shape_name, face_id)

    def pick_face_from_id(self, shape_name: str, id: int):
        self.invoke_method('PickFaceFromId', shape_name,id)

    def pick_face_from_point(self, shape_name: str, x: float, y: float, z: float):
        self.invoke_method('PickFaceFromPoint', shape_name, x, y, z)

    def change_face_id(self, shape_name: str, change_statement: str, version_number: str):
        self.invoke_method('ChangeFaceId', shape_name, change_statement, version_number)

    def change_edge_id(self, shape_name: str, change_statement: str, version_number: str):
        self.invoke_method('ChangeEdgeId', shape_name, change_statement, version_number)

    def change_vertex_id(self, shape_name: str, change_statement: str, version_number: str):
        self.invoke_method('ChangeVertexId', shape_name, change_statement, version_number)

    def delete_face(self, index: int):
        self.invoke_method('DeleteFace', index)

    def delete_point(self, index: int):
        self.invoke_method('DeletePoint', index)

    def add_mean_point(self, indices: list[int]):
        self.invoke_method('MeanPoint', ','.join(str(s) for s in indices))

    def add_mean_from_last_two_points(self):
        self.invoke_method('MeanLastTwoPoints')

    def move_point(self, index: int, dx: float, dy: float, dz: float, keep_original: bool = False):
        self.invoke_method('MovePoint', index, dx, dy, dz, keep_original)

    def pick_point_from_coordinates(self, x: float, y: float, z: float):
        self.invoke_method('PickPointFromCoordinates', x, y, z)

    #TODO: implement missing methods

    def get_edge_id_from_point(self, shape_name: str, x: float, y: float, z: float) -> int:
        return self.invoke_method('GetEdgeIdFromPoint', shape_name, x, y, z)

    def get_face_id_from_point(self, shape_name: str, x: float, y: float, z: float) -> int:
        return self.invoke_method('GetFaceIdFromPoint', shape_name, x, y, z)

    def get_number_of_picked_points(self) -> int:
        return self.invoke_method('GetNumberOfPickedPoints')

    def get_number_of_picked_edges(self) -> int:
        return self.invoke_method('GetNumberOfPickedEdges')

    def get_number_of_picked_facess(self) -> int:
        return self.invoke_method('GetNumberOfPickedFaces')

    #TODO: implement missing methods


from .. import installation_util
from . import ResultModule
from typing import Optional, Union

class Results:
    def __init__(
            self, project_file_path: str,
            version_or_install_dir: Optional[Union[int,str]] = None,
            allow_interactive: bool = True) -> None:
        installation_util.load_win_cst_python_lib(version_or_install_dir)
        import cst.results
        self.proj_file = cst.results.ProjectFile(project_file_path, allow_interactive)

    def get_3d(self) -> ResultModule:
        return ResultModule(self.proj_file.get_3d())

    def get_schematic(self) -> ResultModule:
        return ResultModule(self.proj_file.get_schematic())


from . import Project
from . import ComObjectWrapper

class Background(ComObjectWrapper):
    MATERIAL_NORMAL = 'normal'
    MATERIAL_PEC = 'pec'

    THERMAL_TYPE_NORMAL = 'normal'
    THERMAL_TYPE_PTC = 'ptc'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Background

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_type(self, material_type: str):
        self.invoke_method('Type', material_type)

    def set_permitivity(self, epsilon: float):
        self.invoke_method('Epsilon', epsilon)

    def set_permeability(self, mu: float):
        self.invoke_method('Mu', mu)

    def set_conductivity(self, sigma: float):
        self.invoke_method('ElConductivity', sigma)

    def set_space_x_min(self, value: float):
        self.invoke_method('XminSpace', value)

    def set_space_x_max(self, value: float):
        self.invoke_method('XmaxSpace', value)

    def set_space_x(self, min_value: float, max_value: float):
        self.set_space_x_min(min_value)
        self.set_space_x_max(max_value)

    def set_space_y_min(self, value: float):
        self.invoke_method('YminSpace', value)

    def set_space_y_max(self, value: float):
        self.invoke_method('YmaxSpace', value)

    def set_space_y(self, min_value: float, max_value: float):
        self.set_space_y_min(min_value)
        self.set_space_y_max(max_value)

    def set_space_z_min(self, value: float):
        self.invoke_method('ZminSpace', value)

    def set_space_z_max(self, value: float):
        self.invoke_method('ZmaxSpace', value)

    def set_space_z(self, min_value: float, max_value: float):
        self.set_space_z_min(min_value)
        self.set_space_z_max(max_value)

    def set_thermal_type(self, thermal_type: str):
        self.invoke_method('ThermalType', thermal_type)

    def set_thermal_conductivity(self, value: float):
        self.invoke_method('ThermalConductivity', value)

    def set_apply_in_all_directions(self, flag: bool = True):
        self.invoke_method('ApplyInAllDirections', flag)


from . import Project
from . import ComObjectWrapper

class Component(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Component

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def create(self, component_name: str):
        self.invoke_method('New', component_name)

    def delete(self, component_name: str):
        self.invoke_method('Delete', component_name)

    def delete_all_empty_components(self, component_name: str):
        self.invoke_method('DeleteAllEmptyComponents', component_name)

    def rename(self, old_name: str, new_name: str):
        self.invoke_method('Rename', old_name, new_name)

    def does_exist(self, component_name: str) -> bool:
        return self.invoke_method('DoesExist', component_name)

    def get_next_free_name(self) -> str:
        return self.invoke_method('GetNextFreeName')

    def hide(self, component_name: str):
        self.invoke_method('HideComponent', component_name)

    def show(self, component_name: str):
        self.invoke_method('ShowComponent', component_name)

    def hide_selected(self):
        self.invoke_method('Hide')

    def show_selected(self):
        self.invoke_method('Show')

    def hide_unselected(self):
        self.invoke_method('HideUnselected')

    def show_unselected(self):
        self.invoke_method('ShowUnselected')

    def hide_all(self):
        self.invoke_method('HideAll')

    def show_all(self):
        self.invoke_method('ShowAll')

    def hide_all_ports(self):
        self.invoke_method('HideAllPorts')

    def show_all_ports(self):
        self.invoke_method('ShowAllPorts')

    def hide_all_field_sources(self):
        self.invoke_method('HideAllFieldSources')

    def show_all_field_sources(self):
        self.invoke_method('ShowAllFieldSources')

    def hide_all_lumped_elements(self):
        self.invoke_method('HideAllLumpedElements')

    def show_all_lumped_elements(self):
        self.invoke_method('ShowAllLumpedElements')

    def hide_all_wires(self):
        self.invoke_method('HideAllWires')

    def show_all_wires(self):
        self.invoke_method('ShowAllWires')

    def hide_all_dielectric(self):
        self.invoke_method('HideAllDielectric')

    def show_all_dielectric(self):
        self.invoke_method('ShowAllDielectric')

    def hide_all_metals(self):
        self.invoke_method('HideAllMetals')

    def show_all_metals(self):
        self.invoke_method('ShowAllMetals')


from . import ResultTree

class ResultTreeItem:
    # Types:
    #   _TET indicates tetrahedral
    #   _SRF indicates triangular surface
    #   _TD indicates time monitors
    TYPE_EFIELD_3D = 'Efield3D' # 3D electric field plot
    TYPE_EFIELD_3D_TD = 'Efield3DTD'
    TYPE_EFIELD_3D_TET = 'Efield3D_tet'
    TYPE_EFIELD_3D_SRF = 'Efield3D_srf'
    TYPE_DFIELD_3D = 'Dfield3D' # 3D electric flux plot
    TYPE_DFIELD_3D_TET = 'Dfield3D_tet'
    TYPE_DFIELD_3D_SRF = 'Dfield3D_srf'
    TYPE_HFIELD_3D = 'Hfield3D' # 3D magnetic field plot
    TYPE_HFIELD_3D_TD = 'Hfield3DTD'
    TYPE_HFIELD_3D_TET = 'Hfield3D_tet'
    TYPE_HFIELD_3D_SRF = 'Hfield3D_srf'
    TYPE_BFIELD_3D = 'Bfield3D' # 3D magnetic flux plot
    TYPE_BFIELD_3D_TET = 'Bfield3D_tet'
    TYPE_BFIELD_3D_SRF = 'Bfield3D_srf'
    TYPE_SURFACE_CURRENT = 'SurfaceCurrent' # 3D surface current plot
    TYPE_SURFACE_CURRENT_TD = 'SurfaceCurrentTD'
    TYPE_SURFACE_CURRENT_TET = 'SurfaceCurrent_tet'
    TYPE_SURFACE_CURRENT_SRF = 'SurfaceCurrent_srf'
    TYPE_PFIELD_3D = 'Pfield3D' # 3D power flow plot
    TYPE_PFIELD_3D_TD = 'Pfield3DTD'
    TYPE_PFIELD_3D_TET = 'Pfield3D_tet'
    TYPE_EFIELD_2D = 'Efield2D' # 2D electric field plot
    TYPE_EFIELD_2D_TD = 'Efield2DTD'
    TYPE_HFIELD_2D = 'Hfield2D' # 2D magnetic field plot
    TYPE_HFIELD_2D_TD = 'Hfield2DTD'
    TYPE_PFIELD_2D = 'Pfield2D' # 2D power flow plot
    TYPE_PFIELD_2D_TD = 'Pfield2DTD'
    TYPE_CURRENT_3D = 'Current3D' # Current density
    TYPE_CURRENT_3D_TD = 'Current3DTD'
    TYPE_CURRENT_3D_TET = 'Current3D_tet'
    TYPE_POWER_LOSS_3D = 'PowerLoss3D' # Power loss density
    TYPE_POWER_LOSS_3D_TD = 'PowerLoss3DTD'
    TYPE_POWER_LOSS_3D_TET = 'PowerLoss3D_tet'
    TYPE_EENERGY_3D = 'EEnergy3D' # Electric energy
    TYPE_EENERGY_3D_TD = 'EEnergy3DTD'
    TYPE_EENERGY_3D_TET = 'EEnergy3D_tet'
    TYPE_HENERGY_3D = 'HEnergy3D' # Magnetic energy
    TYPE_HENERGY_3D_TD = 'HEnergy3DTD'
    TYPE_HENERGY_3D_TET = 'HEnergy3D_tet'
    TYPE_SAR_3D = 'SAR3D' # SAR field
    TYPE_FARFIELD = 'Farfield' # Farfield plot
    TYPE_COLOURMAP = 'Colourmap' # Colourmap plot
    TYPE_XY_SIGNAL = 'XYSignal' # General 1D result, needs additional definition by the Subtype method.
    TYPE_NOTEFILE = 'Notefile' # A generic ASCII text file
    TYPE_MACRO = 'Macro' # Visual Basic Macro

    SUBTYPE_COMPLEX = 'Complex'
    SUBTYPE_LINEAR = 'Linear'
    SUBTYPE_DB = 'dB'
    SUBTYPE_PHASE = 'Phase'
    SUBTYPE_TIME = 'Time'
    SUBTYPE_POSITION = 'Position'
    SUBTYPE_ENERGY = 'Energy'
    SUBTYPE_BALANCE = 'Balance'
    SUBTYPE_USER = 'User'

    DELETE_AT_NEVER = 'never'
    DELETE_AT_REBUILD = 'rebuild'
    DELETE_AT_SOLVER_START = 'solverstart'
    DELETE_AT_TRUE_MODEL_CHANGE = 'truemodelchange'

    def __init__(self, result_tree: ResultTree) -> None:
        self.result_tree = result_tree

    def reset(self):
        self.result_tree.invoke_method('Reset')

    def add(self):
        self.result_tree.invoke_method('Add')

    def delete(self):
        self.result_tree.invoke_method('Delete')

    def set_name(self, name: str):
        self.result_tree.invoke_method('Name', name)

    def set_type(self, item_type: str):
        self.result_tree.invoke_method('Type', item_type)

    def set_subtype(self, subtype: str):
        self.result_tree.invoke_method('Subtype', subtype)

    def set_title(self, title: str):
        self.result_tree.invoke_method('Title', title)

    def set_x_label(self, label: str):
        self.result_tree.invoke_method('Xlabel', label)

    def set_y_label(self, label: str):
        self.result_tree.invoke_method('Ylabel', label)

    def set_file(self, file: str):
        self.result_tree.invoke_method('File', file)

    def delete_at(self, when: str):
        self.result_tree.invoke_method('DeleteAt', when)

    def set_as_result(self, flag: bool = True):
        self.result_tree.invoke_method('IsResult', flag)

    def get_result_type(self, item_name: str) -> str:
        return self.result_tree.invoke_method('GetResultTypeFromItemName', item_name)

    def get_file(self, item_name: str) -> str:
        return self.result_tree.invoke_method('GetFileFromTreeItem', item_name)

    def get_result_ids(self, item_name: str):
        return self.result_tree.invoke_method('GetResultIDsFromTreeItem', item_name)

    def get_result(self, item_name: str, result_id: str):
        # TODO: check the result and return corresponding wrapper class
        return self.result_tree.invoke_method('GetResultFromTreeItem', item_name, result_id)

    def get_impedance_result(self, item_name: str, result_id: str):
        # TODO: check the result and return corresponding wrapper class
        return self.result_tree.invoke_method('GetImpedanceResultFromTreeItem', item_name, result_id)

    def has_impedance(self, item_name: str, result_id: str) -> bool:
        return self.result_tree.invoke_method('TreeItemHasImpedance', item_name, result_id)


from . import Project
from . import ComObjectWrapper
from . import w32com
import numpy as np

class FloquetPort(ComObjectWrapper):
    POSITION_Z_MIN = 'zmin'
    POSITION_Z_MAX = 'zmax'

    MODE_TYPE_TE = 'TE'
    MODE_TYPE_TM = 'TM'
    MODE_TYPE_LCP = 'LCP'
    MODE_TYPE_RCP = 'RCP'

    CODE_PLUS_BETA_PW = '+beta/pw'
    CODE_PLUS_BETA = '+beta'
    CODE_MINUS_BETA = '-beta'
    CODE_PLUS_ALPHA = '+alpha'
    CODE_MINUS_ALPHA = '-alpha'
    CODE_PLUS_TE = '+te'
    CODE_MINUS_TE = '-te'
    CODE_PLUS_TM = '+tm'
    CODE_MINUS_TM = '-tm'
    CODE_PLUS_ORDER_X = '+orderx'
    CODE_MINUS_ORDER_X = '-orderx'
    CODE_PLUS_ORDER_Y = '+ordery'
    CODE_MINUS_ORDER_Y = '-ordery'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.FloquetPort

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def prepare_port(self, position: str):
        self.invoke_method('Port', position)

    def add_mode(self, mode_type: str, order_x: int, order_y_prime: int):
        self.invoke_method('AddMode', mode_type, order_x, order_y_prime)

    def set_use_circular_polarization(self, flag: bool = True):
        self.invoke_method('SetUseCircularPolarization', flag)

    def set_polarization_independent_of_phi_deg(self, alignment_angle: float):
        self.invoke_method('SetPolarizationIndependentOfScanAnglePhi', alignment_angle, True)

    def set_polarization_independent_of_phi_rad(self, alignment_angle: float):
        self.set_polarization_independent_of_phi_deg(np.rad2deg(alignment_angle))

    def set_polarization_dependent_of_phi(self):
        self.invoke_method('SetPolarizationIndependentOfScanAnglePhi', 0, False)

    def set_dialog_frequency(self, frequency: float):
        self.invoke_method('SetDialogFrequency', frequency)

    def set_dialog_media_factor(self, frequency: float):
        self.invoke_method('SetDialogMediaFactor', frequency)

    def set_dialog_theta_deg(self, angle: float):
        self.invoke_method('SetDialogTheta', angle)

    def set_dialog_theta_rad(self, angle: float):
        self.set_dialog_theta_deg(np.rad2deg(angle))

    def set_dialog_phi_deg(self, angle: float):
        self.invoke_method('SetDialogPhi', angle)

    def set_dialog_phi_rad(self, angle: float):
        self.set_dialog_phi_deg(np.rad2deg(angle))

    def set_dialog_max_order_x(self, order: int):
        self.invoke_method('SetDialogMaxOrderX', order)

    def set_dialog_max_order_y_prime(self, order: int):
        self.invoke_method('SetDialogMaxOrderYPrime', order)

    def set_customized_list(self, flag: bool = True):
        self.invoke_method('SetCustomizedListFlag', flag)

    def set_number_of_modes_considered(self, number: int):
        self.invoke_method('SetNumberOfModesConsidered', number)

    def set_sort_code(self, code: str):
        self.invoke_method('SetSortCode', code)

    def set_distance_to_reference_plane(self, distance: float):
        self.invoke_method('SetDistanceToReferencePlane', distance)

    def get_number_of_modes(self) -> int:
        return self.invoke_method('GetNumberOfModes')

    def reset_mode_iterator(self) -> bool:
        return self.invoke_method('FirstMode')

    def get_mode(self):
        mode_type = w32com.create_ref_str()
        order_x = w32com.create_ref_int()
        order_y_prime = w32com.create_ref_int()
        success = self.invoke_method('GetMode', mode_type, order_x, order_y_prime)
        return (mode_type.value, order_x.value, order_y_prime.value) if success else None

    def increment_mode_iterator(self) -> bool:
        return self.invoke_method('NextMode')

    def get_number_of_modes_considered(self) -> int:
        return self.invoke_method('GetNumberOfModesConsidered')

    def is_port_at_z_min(self) -> bool:
        return self.invoke_method('IsPortAtZmin')

    def is_port_at_z_max(self) -> bool:
        return self.invoke_method('IsPortAtZmax')

    def get_mode_name_by_number(self, number: int) -> str:
        name = w32com.create_ref_str()
        success = self.invoke_method('GetModeNameByNumber', name, number)
        return name.value if success else None

    def get_mode_name_by_number(self, name: str) -> int:
        number = w32com.create_ref_long()
        success = self.invoke_method('GetModeNumberByName', number, name)
        return number.value if success else None

    def set_force_legacy_phase_reference(self, flag: bool = True):
        self.invoke_method('ForceLegacyPhaseReference', flag)


from . import IVBAProvider, VBAObjWrapper, FrequencyRangeList

class HFSolver(VBAObjWrapper):
    PML_TYPE_CONVOLUTION = 'ConvPML'
    PML_TYPE_GENERALIZED = 'GTPML'

    UPDATE_SCHEMA_GAP = 'Gap'
    UPDATE_SCHEMA_DISTRIBUTED = 'Distributed'

    SURPRESS_TSS_TRUE = 'True'
    SURPRESS_TSS_ALL = 'All'
    SURPRESS_TSS_FALSE = 'False'
    SURPRESS_TSS_NONE = 'None'
    SURPRESS_TSS_RESET = 'Reset'
    SURPRESS_TSS_PORTS = 'Ports'
    SURPRESS_TSS_LUMPED_ELEMENTS = 'LumpedElements'
    SURPRESS_TSS_PROBES = 'Probes'
    SURPRESS_TSS_UI_MONITORS = 'UIMonitors'

    EXCITATION_PORT_MODE = 'portmode'
    EXCITATION_PLANE_WAVE = 'planewave'
    EXCITATION_FIELD_SOURCE = 'fieldsource'
    EXCITATION_FARFIELD_SOURCE = 'farfieldsource'

    SIMULT_EXCIT_OFFSET_TIMESHIFT = 'Timeshift'
    SIMULT_EXCIT_OFFSET_PHASESHIFT = 'Phaseshift'

    ABSORB_AUTOMATIC = 'Automatic'
    ABSORB_ACTIVATE = 'Activate'
    ABSORB_DEACTIVATE = 'Deactivate'

    EXCITATION_TYPE_AUTOMATIC = 'Automatic'
    EXCITATION_TYPE_STANDARD = 'Standard'
    EXCITATION_TYPE_BROADBAND = 'Broadband'

    DECOMPOSITION_TYPE_AUTOMATIC = 'Automatic'
    DECOMPOSITION_TYPE_STANDARD = 'Standard'
    DECOMPOSITION_TYPE_BROADBAND = 'Broadband'

    WINDOW_RECTANGULAR = 'Rectangular'
    WINDOW_COSINE = 'Cosine'

    HFTD_DISP_UPDATE_SCHEME_STANDARD = 'Standard'
    HFTD_DISP_UPDATE_SCHEME_GENERALIZED = 'Generalized'
    HFTD_DISP_UPDATE_SCHEME_AUTOMATIC = 'Automatic'

    SUBCYCLE_STATE_AUTOMATIC = 'Automatic'
    SUBCYCLE_STATE_ACTIVATE = 'Activate'
    SUBCYCLE_STATE_DEACTIVATE = 'Deactivate'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Solver')

    def set_time_between_updates(self, value: int):
        self.record_method('TimeBetweenUpdates', value)

    def set_frequency_range(self, min_freq: float, max_freq: float):
        self.record_method('FrequencyRange', min_freq, max_freq)

    def calculate_zy_matrices(self):
        self.record_method('CalculateZandYMatrices')

    def calculate_vswr(self):
        self.record_method('CalculateVSWR')

    def set_pba_fill_limit(self, percentage: float):
        self.record_method('PBAFillLimit', percentage)

    def use_split_components(self, flag: bool = True):
        self.record_method('UseSplitComponents', flag)

    def always_exclude_pec(self, flag: bool = True):
        self.record_method('AlwaysExcludePec', flag)

    def set_mpi_parallelization(self, flag: bool = True):
        self.record_method('MPIParallelization', flag)

    def get_min_frequency(self) -> float:
        return self.query_method_float('GetFmin')

    def get_max_frequency(self) -> float:
        return self.query_method_float('GetFmax')

    def get_number_of_frequency_samples(self) -> int:
        return self.query_method_int('GetNFsamples')

    def get_number_of_ports(self) -> int:
        return self.query_method_int('GetNumberOfPorts')

    def are_ports_subsequently_named(self) -> bool:
        return self.query_method_bool('ArePortsSubsequentlyNamed')

    def get_stimulation_port(self) -> int:
        return self.query_method_int('GetStimulationPort')

    def get_stimulation_mode(self) -> int:
        return self.query_method_int('GetStimulationMode')

    def get_total_simulation_time(self) -> int:
        return self.query_method_int('GetTotalSimulationTime')

    def get_matrix_calculation_time(self) -> int:
        return self.query_method_int('GetMatrixCalculationTime')

    def get_last_solver_time(self) -> int:
        return self.query_method_int('GetLastSolverTime')

    def start(self):
        self.record_method('Start')

    def set_normalize_impedances_automatically(self, flag: bool = True):
        self.record_method('AutoNormImpedance', flag)

    def set_norming_impedance(self, impedance: float):
        self.record_method('NormingImpedance', impedance)

    def set_mesh_adaptation(self, flag: bool = True):
        self.record_method('MeshAdaption', flag)

    def set_use_distributed_computing(self, flag: bool = True):
        self.record_method('UseDistributedComputing', flag)

    def set_distribute_matrix_calculation(self, flag: bool = True):
        self.record_method('DistributeMatrixCalculation', flag)

    def set_use_hardware_acceleration(self, flag: bool = True):
        self.record_method('HardwareAcceleration', flag)

    def set_store_td_results_in_cache(self, flag: bool = True):
        self.record_method('StoreTDResultsInCache', flag)

    def set_number_of_frequency_samples(self, num_samples: int):
        self.record_method('FrequencySamples', num_samples)

    def set_number_of_log_frequency_samples(self, num_samples: int):
        self.record_method('FrequencyLogSamples', num_samples)

    def set_consider_two_port_reciprocity(self, flag: bool = True):
        self.record_method('ConsiderTwoPortReciprocity', flag)

    def set_energy_balance_limit(self, limit: float):
        self.record_method('EnergyBalanceLimit', limit)

    def set_time_step_stability_factor(self, factor: float):
        self.record_method('TimeStepStabilityFactor', factor)

    def set_normalize_to_reference_signal(self, flag: bool = True):
        self.record_method('NormalizeToReferenceSignal', flag)

    def set_normalize_to_default_signal_when_in_use(self, flag: bool = True):
        self.record_method('NormalizeToDefaultSignalWhenInUse', flag)

    def set_detect_identical_ports_automatically(self, flag: bool = True):
        self.record_method('AutoDetectIdenticalPorts', flag)

    def set_sample_time_signal_automatically(self, flag: bool = True):
        self.record_method('AutomaticTimeSignalSampling', flag)

    def set_consider_excitation_for_freq_sampling_rate(self, flag: bool = True):
        self.record_method('ConsiderExcitationForFreqSamplingRate', flag)

    def set_perform_tdr_computation(self, flag: bool = True):
        self.record_method('TDRComputation', flag)

    def set_shift_tdr_50_percent(self, flag: bool = True):
        self.record_method('TDRShift50Percent', flag)

    def set_perform_tdr_reflection_computation(self, flag: bool = True):
        self.record_method('TDRReflection', flag)

    def set_use_broadband_phase_shift(self, flag: bool = True):
        self.record_method('UseBroadBandPhaseShift', flag)

    def set_broadband_phase_shift_lower_bound_factor(self, value: float):
        self.record_method('SetBroadBandPhaseShiftLowerBoundFac', value)

    def set_sparam_adjustment(self, flag: bool = True):
        self.record_method('SParaAdjustment', flag)

    def set_prepare_farfields(self, flag: bool = True):
        self.record_method('PrepareFarfields', flag)

    def set_monitor_farfields_near_to_model(self, flag: bool = True):
        self.record_method('MonitorFarFieldsNearToModel', flag)

    def set_dump_solver_matrices(self, flag: bool = True):
        self.record_method('MatrixDump', flag)

    def set_restart_after_instability_abort(self, flag: bool = True):
        self.record_method('RestartAfterInstabilityAbort', flag)

    def set_max_number_of_threads(self, num_threads: int):
        self.record_method('MaximumNumberOfThreads', num_threads)

    def set_pml_type(self, pml_type: str):
        self.record_method('SetPMLType', pml_type)

    def set_update_schema_discrete_items(self, schema: str):
        self.record_method('DiscreteItemUpdate', schema)

    def set_update_schema_discrete_items_edge(self, schema: str):
        self.record_method('DiscreteItemEdgeUpdate', schema)

    def set_update_schema_discrete_items_face(self, schema: str):
        self.record_method('DiscreteItemFaceUpdate', schema)

    def get_update_schema_discrete_items(self) -> str:
        return self.query_method_str('GetDiscreteItemUpdate')

    def get_update_schema_discrete_items_edge(self) -> str:
        return self.query_method_str('GetDiscreteItemEdgeUpdate')

    def get_update_schema_discrete_items_face(self) -> str:
        return self.query_method_str('GetDiscreteItemFaceUpdate')

    def surpress_time_signal_storage(self, key: str):
        self.record_method('SuppressTimeSignalStorage', key)

    def set_calculate_modes_only(self, flag: bool = True):
        self.record_method('CalculateModesOnly', flag)

    def set_stimulation_mode(self, mode_number: int):
        self.record_method('StimulationMode', mode_number)

    def set_stimulation_by_all_ports(self):
        self.record_method('StimulationPort', 'All')

    def set_stimulation_by_selected_port(self):
        self.record_method('StimulationPort', 'Selected')

    def set_stimulation_by_plane_wave(self):
        self.record_method('StimulationPort', 'Plane Wave')

    def set_stimulation_by_port(self, port_number: int):
        self.record_method('StimulationPort', port_number)

    def reset_excitation_modes(self):
        self.record_method('ResetExcitationModes')

    def set_excitation_port_mode(
            self, port: int, mode: int, amplitude: float, phase: float,
            signal: str = "default", active: bool = True):
        self.record_method('ExcitationPortMode', port, mode, amplitude, phase, signal, active)

    def set_excitation_field_source(
            self, name: str, amplitude: float, phase: float,
            signal: str = "default", active: bool = True):
        self.record_method('ExcitationFieldSource', name, amplitude, phase, signal, active)

    def define_excitation(
            self, excitation: str, name: int, mode: int, amplitude: float, phase_or_time:
            float, signal_name: str, active: bool = True):
        self.record_method(
            'DefineExcitation', excitation, name, mode, amplitude, phase_or_time,
            signal_name, active)

    def define_excitation_settings(
            self, excitation: str, name: int, mode: int, amplitude: float, phase_or_time:
            float, signal_name: str, accuracy: float, f_min: float, f_max: float,
            active: bool = True):
        self.record_method(
            'DefineExcitationSettings', excitation, name, mode, amplitude, phase_or_time,
            signal_name, accuracy, f_min, f_max, active)

    def set_phase_ref_frequency(self, frequency: float):
        self.record_method('PhaseRefFrequency', frequency)

    def set_s_param_port_excitation(self, flag: bool = True):
        self.record_method('SParameterPortExcitation', flag)

    def set_simultaneous_excitation(self, flag: bool = True):
        self.record_method('SimultaneousExcitation', flag)

    def set_simultaneous_excitation_auto_label(self, flag: bool = True):
        self.record_method('SetSimultaneousExcitAutoLabel', flag)

    def set_simultaneous_excitation_label(self, label: str):
        self.record_method('SetSimultaneousExcitationLabel', label)

    def set_simultaneous_excitation_offset(self, offset: str):
        self.record_method('SetSimultaneousExcitationOffset', offset)

    def set_excitation_selection_show_additional_settings(self, flag: bool = True):
        self.record_method('ExcitationSelectionShowAdditionalSettings', flag)

    def set_superimpose_plw_excitation(self, flag: bool = True):
        self.record_method('SuperimposePLWExcitation', flag)

    def set_show_excitation_list_accuracy(self, flag: bool = True):
        self.record_method('ShowExcitationListAccuracy', flag)

    def set_show_excitation_list_monitor_freq_interval(self, flag: bool = True):
        self.record_method('ShowExcitationListMonitorFreqInterval', flag)

    def set_use_s_param_symmetries(self, flag: bool = True):
        self.record_method('SParaSymmetry', flag)

    def reset_s_param_symmetries(self):
        self.record_method('ResetSParaSymm')

    def define_new_s_param_symmetries(self):
        self.record_method('DefSParaSymm')

    def add_s_param(self, port_num1: int, port_num2: int):
        self.record_method('SPara', port_num1, port_num2)

    def set_waveguide_port_generalized(self, flag: bool = True):
        self.record_method('WaveguidePortGeneralized', flag)

    def set_waveguide_port_mode_tracking(self, flag: bool = True):
        self.record_method('WaveguidePortModeTracking', flag)

    def set_absorb_unconsidered_mode_fields(self, key: str):
        self.record_method('AbsorbUnconsideredModeFields', key)

    def set_absorb_unconsidered_mode_fields(self, flag: bool = True):
        key = self.ABSORB_ACTIVATE if flag else self.ABSORB_DEACTIVATE
        self.record_method('AbsorbUnconsideredModeFields', key)

    def set_absorb_unconsidered_mode_fields_auto(self):
        self.record_method('AbsorbUnconsideredModeFields', self.ABSORB_AUTOMATIC)

    def set_full_deembedding(self, flag: bool = True):
        self.record_method('FullDeembedding', flag)

    def set_number_of_samples_full_deembedding(self, number: int):
        self.record_method('SetSamplesFullDeembedding', number)

    def set_consider_dispersive_behavior_full_deembedding(self, flag: bool = True):
        self.record_method('DispEpsFullDeembedding', flag)

    def set_waveguide_port_broadband(self, flag: bool = True):
        self.record_method('WaveguidePortBroadband', flag)

    def set_mode_freq_factor(self, factor: float):
        self.record_method('SetModeFreqFactor', factor)

    def set_scale_te_tm_mode_to_center_frequency(self, flag: bool = True):
        self.record_method('ScaleTETMModeToCenterFrequency', flag)

    def set_waveguide_port_excitation_type(self, excitation_type: str):
        self.record_method('WaveguidePortExcitationType', excitation_type)

    def set_waveguide_port_decomposition_type(self, decomposition_type: str):
        self.record_method('WaveguidePortDecompositionType', decomposition_type)

    def set_voltage_waveguide_port(self, port_name: int, value: float):
        self.record_method('SetVoltageWaveguidePort', port_name, value, True)

    def unset_voltage_waveguide_port(self, port_name: int):
        self.record_method('SetVoltageWaveguidePort', port_name, 0.0, False)

    def set_adaptive_port_meshing(self, flag: bool = True):
        self.record_method('AdaptivePortMeshing', flag)

    def set_adaptive_port_meshing_accuracy(self, accuracy_percent: int):
        self.record_method('AccuracyAdaptivePortMeshing', accuracy_percent)

    def set_adaptive_port_meshing_number_of_passes(self, number: int):
        self.record_method('PassesAdaptivePortMeshing', number)

    def set_number_of_pulse_widths(self, number: int):
        self.record_method('NumberOfPulseWidths', number)

    def enable_steady_state_monitor(self, accuracy_dB: float):
        self.record_method('SteadyStateLimit', accuracy_dB)

    def disable_steady_state_monitor(self):
        self.record_method('SteadyStateLimit', 'No Check')

    def add_stop_criterion(
            self, group_name: str, threshold: float, num_checks: int, active: bool = True):
        self.record_method('AddStopCriterion', group_name, threshold, num_checks, active)

    def add_stop_criterion_with_target_frequency(
            self, group_name: str, threshold: float, num_checks: int,
            target_frequency_ranges: FrequencyRangeList, active: bool = True):
        self.record_method(
            'AddStopCriterionWithTargetFrequency', group_name, threshold, num_checks,
            active, str(target_frequency_ranges))

    def set_stop_criteria_show_excitation(self, flag: bool = True):
        self.record_method('StopCriteriaShowExcitation', flag)

    def set_use_ar_filter(self, flag: bool = True):
        self.record_method('UseArfilter', flag)

    def set_ar_max_energy_deviation(self, limit: float):
        self.record_method('ArMaxEnergyDeviation', limit)

    def set_ar_number_of_pulses_to_skip(self, number: int):
        self.record_method('ArPulseSkip', number)

    def start_ar_filter(self):
        self.record_method('StartArFilter')

    def set_time_window(
            self, window: str, smoothness_percent: int, enabled_on_transient_sim: bool = True):
        self.record_method('SetTimeWindow', window, smoothness_percent, enabled_on_transient_sim)

    def set_surface_impedance_order(self, order: int):
        self.record_method('SurfaceImpedanceOrder', order)

    def set_activate_power_loss_1d_monitor(self, flag: bool = True):
        self.record_method('ActivatePowerLoss1DMonitor', flag)

    def set_power_loss_1d_monitor_per_solid(self, flag: bool = True):
        self.record_method('PowerLoss1DMonitorPerSolid', flag)

    def set_use_3d_field_monitor_for_power_loss_1d_monitor(self, flag: bool = True):
        self.record_method('Use3DFieldMonitorForPowerLoss1DMonitor', flag)

    def set_use_farfield_monitor_for_power_loss_1d_monitor(self, flag: bool = True):
        self.record_method('UseFarFieldMonitorForPowerLoss1DMonitor', flag)

    def set_use_extra_freq_for_power_loss_1d_monitor(self, flag: bool = True):
        self.record_method('UseExtraFreqForPowerLoss1DMonitor', flag)

    def reset_power_loss_1d_monitor_extra_freqs(self):
        self.record_method('ResetPowerLoss1DMonitorExtraFreq')

    def add_power_loss_1d_monitor_extra_freq(self, freq: float):
        self.record_method('AddPowerLoss1DMonitorExtraFreq', freq)

    def set_time_power_loss_si_material_monitor(self, flag: bool = True):
        self.record_method('SetTimePowerLossSIMaterialMonitor', flag)

    def activate_timer_power_loss_si_material_monitor(self, t_start: float, t_step: float):
        self.record_method('ActivateTimePowerLossSIMaterialMonitor', t_start, t_step, 0.0, False)

    def activate_timer_power_loss_si_material_monitor_lim(
            self, t_start: float, t_step: float, t_end: float):
        self.record_method('ActivateTimePowerLossSIMaterialMonitor', t_start, t_step, t_end, True)

    def set_time_power_loss_si_material_monitor_average(self, flag: bool = True):
        self.record_method('SetTimePowerLossSIMaterialMonitorAverage', flag)

    def set_time_power_loss_si_material_monitor_average_rep_period(self, period: float):
        self.record_method('SetTimePowerLossSIMaterialMonitorAverageRepPeriod', period)

    def set_time_power_loss_si_material_monitor_per_solid(self, flag: bool = True):
        self.record_method('TimePowerLossSIMaterialMonitorPerSolid', flag)

    def set_disp_non_linear_material_monitor(self, flag: bool = True):
        self.record_method('SetDispNonLinearMaterialMonitor', flag)

    def activate_disp_non_linear_material_monitor(self, t_start: float, t_step: float):
        self.record_method('ActivateDispNonLinearMaterialMonitor', t_start, t_step, 0.0, False)

    def activate_disp_non_linear_material_monitor_lim(
            self, t_start: float, t_step: float, t_end: float):
        self.record_method('ActivateDispNonLinearMaterialMonitor', t_start, t_step, t_end, True)

    def set_activate_space_material_3d_monitor(self, flag: bool = True):
        self.record_method('ActivateSpaceMaterial3DMonitor', flag)

    def set_use_3d_field_monitor_for_space_material_3d_monitor(self, flag: bool = True):
        self.record_method('Use3DFieldMonitorForSpaceMaterial3DMonitor', flag)

    def set_use_extra_freq_for_space_material_3d_monitor(self, flag: bool = True):
        self.record_method('UseExtraFreqForSpaceMaterial3DMonitor', flag)

    def reset_space_material_3d_monitor_extra_freqs(self):
        self.record_method('ResetSpaceMaterial3DMonitorExtraFreq')

    def add_space_material_3d_monitor_extra_freq(self, freq: float):
        self.record_method('AddSpaceMaterial3DMonitorExtraFreq', freq)

    def set_hftd_disp_update_scheme(self, scheme: str):
        self.record_method('SetHFTDDispUpdateScheme', scheme)

    def set_use_tst_at_port(self, flag: bool = True):
        self.record_method('UseTSTAtPort', flag)

    def set_sybcycle_state(self, state: str):
        self.record_method('SetSubcycleState', state)

    def set_simplified_pba_method(self, flag: bool = True):
        self.record_method('SimplifiedPBAMethod', flag)

    def set_aks_penalty_factor(self, factor: float):
        self.record_method('AKSPenaltyFactor', factor)

    def set_aks_estimation(self, factor: float):
        self.record_method('AKSEstimation', factor)

    def set_aks_number_of_iterations(self, number: int):
        self.record_method('AKSIterations', number)

    def set_aks_accuracy(self, accuracy: float):
        self.record_method('AKSAccuracy', accuracy)

    def reset_aks(self):
        self.record_method('AKSReset')

    def start_aks(self):
        self.record_method('AKSStart')

    def set_aks_number_of_estimation_cycles(self, number: int):
        self.record_method('AKSEstimationCycles', number)

    def set_aks_automatic_estimation(self, flag: bool = True):
        self.record_method('AKSAutomaticEstimation', flag)

    def set_aks_number_of_check_modes(self, number: int):
        self.record_method('AKSCheckModes', number)

    def set_aks_maximum_df(self, delta_frequency: float):
        self.record_method('AKSMaximumDF', delta_frequency)

    def set_aks_max_number_of_passes(self, number: int):
        self.record_method('AKSMaximumPasses', number)

    def set_aks_mesh_increment(self, inc: int):
        self.record_method('AKSMeshIncrement', inc)

    def set_aks_min_number_of_passes(self, number: int):
        self.record_method('AKSMinimumPasses', number)

    def get_aks_number_of_modes(self) -> int:
        return self.query_method_int('AKSGetNumberOfModes')


from . import IVBAProvider, VBAObjWrapper, VBATypeName
import numpy as np

class Boundary(VBAObjWrapper):
    BOUNDARY_ELECTRIC = 'electric'
    BOUNDARY_MAGNETIC = 'magnetic'
    BOUNDARY_TANGENTIAL = 'tangential'
    BOUNDARY_NORMAL = 'normal'
    BOUNDARY_OPEN = 'open'
    BOUNDARY_EXPANDED_OPEN = 'expanded open'
    BOUNDARY_PERIODIC = 'periodic'
    BOUNDARY_CONDUCTING_WALL = 'conducting wall'
    BOUNDARY_UNIT_CELL = 'unit cell'

    SYMMETRY_ELECTRIC = 'electric'
    SYMMETRY_MAGNETIC = 'magnetic'
    SYMMETRY_NONE = 'none'

    POTENTIAL_NONE = 'none'
    POTENTIAL_FIXED = 'fixed'
    POTENTIAL_FLOATING = 'floating'

    THERMAL_BOUNDARY_ISOTHERMAL = 'isothermal'
    THERMAL_BOUNDARY_ADIABATIC = 'adiabatic'
    THERMAL_BOUNDARY_OPEN = 'open'
    THERMAL_BOUNDARY_EXPANDED_OPEN = 'expanded open'

    THERMAL_SYMMETRY_ISOTHERMAL = 'isothermal'
    THERMAL_SYMMETRY_ADIABATIC = 'adiabatic'
    THERMAL_SYMMETRY_NONE = 'none'
    THERMAL_SYMMETRY_EXPANDED_OPEN = 'expanded open'

    TEMPERATURE_NONE = 'none'
    TEMPERATURE_FIXED = 'fixed'
    TEMPERATURE_FLOATING = 'floating'

    DISTANCE_FRACTION = 'Fraction'
    DISTANCE_ABSOLUTE = 'Absolute'

    DISTANCE_REF_FREQ_CENTER = 'Center'
    DISTANCE_REF_FREQ_CENTER_AND_MONITORS = 'CenterNMonitors'
    DISTANCE_REF_FREQ_USER = 'User'

    DIRECTION_OUTWARD = 'outward'
    DIRECTION_INWARD = 'inward'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Boundary')

    def set_type_x_min(self, boundary_type: str):
        self.record_method('Xmin', boundary_type)

    def set_type_x_max(self, boundary_type: str):
        self.record_method('Xmax', boundary_type)

    def set_types_x(self, min_boudary_type: str, max_boudary_type: str):
        self.set_type_x_min(min_boudary_type)
        self.set_type_x_max(max_boudary_type)

    def set_type_y_min(self, boundary_type: str):
        self.record_method('Ymin', boundary_type)

    def set_type_y_max(self, boundary_type: str):
        self.record_method('Ymax', boundary_type)

    def set_types_y(self, min_boudary_type: str, max_boudary_type: str):
        self.set_type_y_min(min_boudary_type)
        self.set_type_y_max(max_boudary_type)

    def set_type_z_min(self, boundary_type: str):
        self.record_method('Zmin', boundary_type)

    def set_type_z_max(self, boundary_type: str):
        self.record_method('Zmax', boundary_type)

    def set_types_z(self, min_boudary_type: str, max_boudary_type: str):
        self.set_type_z_min(min_boudary_type)
        self.set_type_z_max(max_boudary_type)

    def set_types(self, x_min: str, x_max: str, y_min: str, y_max: str, z_min: str, z_max: str):
        self.set_types_x(x_min, x_max)
        self.set_types_y(y_min, y_max)
        self.set_types_z(z_min, z_max)

    def get_type_x_min(self) -> str:
        return self.query_method_str('GetXmin')

    def get_type_x_max(self) -> str:
        return self.query_method_str('GetXmax')

    def get_type_y_min(self) -> str:
        return self.query_method_str('GetYmin')

    def get_type_y_max(self) -> str:
        return self.query_method_str('GetYmax')

    def get_type_z_min(self) -> str:
        return self.query_method_str('GetZmin')

    def get_type_z_max(self) -> str:
        return self.query_method_str('GetZmax')

    def set_symmetry_x(self, symmetry_type: str):
        self.record_method('Xsymmetry', symmetry_type)

    def set_symmetry_y(self, symmetry_type: str):
        self.record_method('Ysymmetry', symmetry_type)

    def set_symmetry_z(self, symmetry_type: str):
        self.record_method('Zsymmetry', symmetry_type)

    def set_symmetries(self, x_type: str, y_type: str, z_type: str):
        self.set_symmetry_x(x_type)
        self.set_symmetry_y(y_type)
        self.set_symmetry_z(z_type)

    def get_symmetry_x(self) -> str:
        return self.query_method_str('GetXSymmetry')

    def get_symmetry_y(self) -> str:
        return self.query_method_str('GetYSymmetry')

    def get_symmetry_z(self) -> str:
        return self.query_method_str('GetZSymmetry')

    def set_apply_in_all_directions(self, flag: bool = True):
        self.record_method('ApplyInAllDirections', flag)

    def set_potential_type_x_min(self, potential_type: str):
        self.record_method('XminPotentialType', potential_type)

    def set_potential_type_x_max(self, potential_type: str):
        self.record_method('XmaxPotentialType', potential_type)

    def set_potential_types_x(self, min_type: str, max_type: str):
        self.set_potential_type_x_min(min_type)
        self.set_potential_type_x_max(max_type)

    def set_potential_type_y_min(self, potential_type: str):
        self.record_method('YminPotentialType', potential_type)

    def set_potential_type_y_max(self, potential_type: str):
        self.record_method('YmaxPotentialType', potential_type)

    def set_potential_types_y(self, min_type: str, max_type: str):
        self.set_potential_type_y_min(min_type)
        self.set_potential_type_y_max(max_type)

    def set_potential_type_z_min(self, potential_type: str):
        self.record_method('ZminPotentialType', potential_type)

    def set_potential_type_z_max(self, potential_type: str):
        self.record_method('ZmaxPotentialType', potential_type)

    def set_potential_types_z(self, min_type: str, max_type: str):
        self.set_potential_type_z_min(min_type)
        self.set_potential_type_z_max(max_type)

    def set_potential_types(self, x_min: str, x_max: str, y_min: str, y_max: str, z_min: str, z_max: str):
        self.set_potential_types_x(x_min, x_max)
        self.set_potential_types_y(y_min, y_max)
        self.set_potential_types_z(z_min, z_max)

    def set_potential_x_min(self, potential: float):
        self.record_method('XminPotential', potential)

    def set_potential_x_max(self, potential: float):
        self.record_method('XmaxPotential', potential)

    def set_potentials_x(self, min_potential: float, max_potential: float):
        self.set_potential_x_min(min_potential)
        self.set_potential_x_max(max_potential)

    def set_potential_y_min(self, potential: float):
        self.record_method('YminPotential', potential)

    def set_potential_y_max(self, potential: float):
        self.record_method('YmaxPotential', potential)

    def set_potentials_y(self, min_potential: float, max_potential: float):
        self.set_potential_y_min(min_potential)
        self.set_potential_y_max(max_potential)

    def set_potential_z_min(self, potential: float):
        self.record_method('ZminPotential', potential)

    def set_potential_z_max(self, potential: float):
        self.record_method('ZmaxPotential', potential)

    def set_potentials_z(self, min_potential: float, max_potential: float):
        self.set_potential_z_min(min_potential)
        self.set_potential_z_max(max_potential)

    def set_potentials(
            self, x_min: float, x_max: float, y_min: float, y_max: float,
            z_min: float, z_max: float):
        self.set_potentials_x(x_min, x_max)
        self.set_potentials_y(y_min, y_max)
        self.set_potentials_z(z_min, z_max)

    def set_thermal_type_x_min(self, thermal_type: str):
        self.record_method('XminThermal', thermal_type)

    def set_thermal_type_x_max(self, thermal_type: str):
        self.record_method('XmaxThermal', thermal_type)

    def set_thermal_types_x(self, min_type: str, max_type: str):
        self.set_thermal_type_x_min(min_type)
        self.set_thermal_type_x_max(max_type)

    def set_thermal_type_y_min(self, thermal_type: str):
        self.record_method('YminThermal', thermal_type)

    def set_thermal_type_y_max(self, thermal_type: str):
        self.record_method('YmaxThermal', thermal_type)

    def set_thermal_types_y(self, min_type: str, max_type: str):
        self.set_thermal_type_y_min(min_type)
        self.set_thermal_type_y_max(max_type)

    def set_thermal_type_z_min(self, thermal_type: str):
        self.record_method('ZminThermal', thermal_type)

    def set_thermal_type_z_max(self, thermal_type: str):
        self.record_method('ZmaxThermal', thermal_type)

    def set_thermal_types_z(self, min_type: str, max_type: str):
        self.set_thermal_type_z_min(min_type)
        self.set_thermal_type_z_max(max_type)

    def set_thermal_types(self, x_min: str, x_max: str, y_min: str, y_max: str, z_min: str, z_max: str):
        self.set_thermal_types_y(y_min, y_max)
        self.set_thermal_types_x(x_min, x_max)
        self.set_thermal_types_z(z_min, z_max)

    def set_thermal_symmetry_x(self, symmetry_type: str):
        self.record_method('XsymmetryThermal', symmetry_type)

    def set_thermal_symmetry_y(self, symmetry_type: str):
        self.record_method('YsymmetryThermal', symmetry_type)

    def set_thermal_symmetry_z(self, symmetry_type: str):
        self.record_method('ZsymmetryThermal', symmetry_type)

    def set_thermal_symmetries(self, x: str, y: str, z: str):
        self.set_thermal_symmetry_x(x)
        self.set_thermal_symmetry_y(y)
        self.set_thermal_symmetry_z(z)

    def get_thermal_symmetry_x(self) -> str:
        return self.query_method_str('GetXSymmetryThermal')

    def get_thermal_symmetry_y(self) -> str:
        return self.query_method_str('GetYSymmetryThermal')

    def get_thermal_symmetry_z(self) -> str:
        return self.query_method_str('GetZSymmetryThermal')

    def set_temperature_type_x_min(self, temperature_type: str):
        self.record_method('XminTemperatureType', temperature_type)

    def set_temperature_type_x_max(self, temperature_type: str):
        self.record_method('XmaxTemperatureType', temperature_type)

    def set_temperature_types_x(self, min_type: str, max_type: str):
        self.set_temperature_type_x_min(min_type)
        self.set_temperature_type_x_max(max_type)

    def set_temperature_type_y_min(self, temperature_type: str):
        self.record_method('YminTemperatureType', temperature_type)

    def set_temperature_type_y_max(self, temperature_type: str):
        self.record_method('YmaxTemperatureType', temperature_type)

    def set_temperature_types_y(self, min_type: str, max_type: str):
        self.set_temperature_type_y_min(min_type)
        self.set_temperature_type_y_max(max_type)

    def set_temperature_type_z_min(self, temperature_type: str):
        self.record_method('ZminTemperatureType', temperature_type)

    def set_temperature_type_z_max(self, temperature_type: str):
        self.record_method('ZmaxTemperatureType', temperature_type)

    def set_temperature_types_z(self, min_type: str, max_type: str):
        self.set_temperature_type_z_min(min_type)
        self.set_temperature_type_z_max(max_type)

    def set_temperature_types(
            self, x_min: str, x_max: str, y_min: str, y_max: str, z_min: str, z_max: str):
        self.set_temperature_types_x(x_min, x_max)
        self.set_temperature_types_y(y_min, y_max)
        self.set_temperature_types_z(z_min, z_max)

    def set_temperature_x_min(self, temperature: float):
        self.record_method('XminTemperature', temperature)

    def set_temperature_x_max(self, temperature: float):
        self.record_method('XmaxTemperature', temperature)

    def set_temperatures_x(self, min_temperature: float, max_temperature: float):
        self.set_temperature_x_min(min_temperature)
        self.set_temperature_x_max(max_temperature)

    def set_temperature_y_min(self, temperature: float):
        self.record_method('YminTemperature', temperature)

    def set_temperature_y_max(self, temperature: float):
        self.record_method('YmaxTemperature', temperature)

    def set_temperatures_y(self, min_temperature: float, max_temperature: float):
        self.set_temperature_y_min(min_temperature)
        self.set_temperature_y_max(max_temperature)

    def set_temperature_z_min(self, temperature: float):
        self.record_method('ZminTemperature', temperature)

    def set_temperature_z_max(self, temperature: float):
        self.record_method('ZmaxTemperature', temperature)

    def set_temperatures_z(self, min_temperature: float, max_temperature: float):
        self.set_temperature_z_min(min_temperature)
        self.set_temperature_z_max(max_temperature)

    def set_temperatures(
            self, x_min: float, x_max: float, y_min: float, y_max: float,
            z_min: float, z_max: float):
        self.set_temperatures_x(x_min, x_max)
        self.set_temperatures_y(y_min, y_max)
        self.set_temperatures_z(z_min, z_max)

    # returns: (x_min, x_max, y_min, y_max, z_min, z_max)
    def get_calculation_box(self) -> tuple[float, float, float, float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetCalculationBox', None, D, D, D, D, D, D)

    # returns: (x_min, x_max, y_min, y_max, z_min, z_max)
    def get_structure_box(self) -> tuple[float, float, float, float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetStructureBox', None, D, D, D, D, D, D)

    def set_number_of_layers(self, count: int):
        self.record_method('Layer', count)

    def set_minimum_lines_distance(self, distance: float):
        self.record_method('MinimumLinesDistance', distance)

    def set_minimum_distance_type(self, distance_type: str):
        self.record_method('MinimumDistanceType', distance_type)

    def set_absolute_distance(self, distance: float):
        self.record_method('SetAbsoluteDistance', distance)

    def set_minimum_distance_reference_frequency_type(self, freq_type: str):
        self.record_method('MinimumDistanceReferenceFrequencyType', freq_type)

    def set_minimum_distance_per_wavelength(self, distance: float):
        self.record_method('MinimumDistancePerWavelength', distance)

    def set_frequency_for_minimum_distance(self, freq: float):
        self.record_method('FrequencyForMinimumDistance', freq)

    def set_period_shift_x_deg(self, angle_deg: float):
        self.record_method('XPeriodicShift', angle_deg)

    def set_period_shift_x_rad(self, angle_rad: float):
        self.set_period_shift_x_deg(np.rad2deg(angle_rad))

    def set_period_shift_y_deg(self, angle_deg: float):
        self.record_method('YPeriodicShift', angle_deg)

    def set_period_shift_y_rad(self, angle_rad: float):
        self.set_period_shift_y_deg(np.rad2deg(angle_rad))

    def set_period_shift_z_deg(self, angle_deg: float):
        self.record_method('ZPeriodicShift', angle_deg)

    def set_period_shift_z_rad(self, angle_rad: float):
        self.set_period_shift_z_deg(np.rad2deg(angle_rad))

    def set_periodic_use_constant_angles(self, flag: bool = True):
        self.record_method('PeriodicUseConstantAngles', flag)

    def set_periodic_boundary_angles_deg(self, theta_deg: float, phi_deg: float):
        self.record_method('SetPeriodicBoundaryAngles', theta_deg, phi_deg)

    def set_periodic_boundary_angles_rad(self, theta_rad: float, phi_rad: float):
        self.set_periodic_boundary_angles_deg(np.rad2deg(theta_rad), np.rad2deg(phi_rad))

    def set_periodic_boundary_angles_direction(self, direction: str):
        self.record_method('SetPeriodicBoundaryAnglesDirection', direction)

    # returns: (theta, phi, dir)
    def get_unit_cell_scan_angle(self) -> tuple[float, float, str]:
        D, L = VBATypeName.Double, VBATypeName.Long
        res = self.query_method_t('GetUnitCellScanAngle', None, D, D, L)
        dir_str = self.DIRECTION_OUTWARD if res[2] == 1 else self.DIRECTION_INWARD
        return (res[0], res[1], dir_str)

    def set_unit_cells_distance_dir1(self, distance: float):
        self.record_method('UnitCellDs1', distance)

    def set_unit_cells_distance_dir2(self, distance: float):
        self.record_method('UnitCellDs2', distance)

    def get_unit_cells_distance_dir1(self) -> float:
        return self.query_method_float('GetUnitCellDs1')

    def get_unit_cells_distance_dir2(self) -> float:
        return self.query_method_float('GetUnitCellDs2')

    def set_unit_cells_angle_deg(self, angle_deg: float):
        self.record_method('UnitCellAngle', angle_deg)

    def set_unit_cells_angle_rad(self, angle_rad: float):
        self.set_unit_cells_angle_deg(np.rad2deg(angle_rad))

    def get_unit_cells_angle_deg(self) -> float:
        return self.query_method_float('UnitCellAngle')

    def get_unit_cells_angle_rad(self) -> float:
        return np.deg2rad(self.get_unit_cells_angle_deg())

    def set_unit_cells_origin(self, pos_dir1: float, pos_dir2: float):
        self.record_method('UnitCellOrigin', pos_dir1, pos_dir2)

    def set_unit_cells_fit_to_bounding_box(self, flag: bool = True):
        self.record_method('UnitCellFitToBoundingBox', flag)


from . import Project
from . import ComObjectWrapper

class Shape(ComObjectWrapper):
    def __init__(self, project: Project, com_object) -> None:
        self.project = project
        self.com_object = com_object

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_component(self, component_name: str):
        self.invoke_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.invoke_method('Material', mat_name)

    def create(self):
        self.invoke_method('Create')


from . import Project
from . import ComObjectWrapper
from . import w32com

class DiscretePort(ComObjectWrapper):
    TYPE_S_PARAMETER = 'Sparameter'
    TYPE_VOLTAGE = 'Voltage'
    TYPE_CURRENT = 'Current'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.DiscretePort

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def create(self):
        self.invoke_method('Create')

    def modify(self):
        self.invoke_method('Modify')

    def set_number(self, number: int):
        self.invoke_method('PortNumber', number)

    def set_label(self, label: str):
        self.invoke_method('Label', label)

    def set_type(self, port_type: str):
        self.invoke_method('Type', port_type)

    def set_impedance(self, impedance: float):
        self.invoke_method('Impedance', impedance)

    def set_voltage(self, voltage: float):
        self.invoke_method('Voltage', voltage)

    def set_voltage_port_impedance(self, impedance: float):
        self.invoke_method('VoltagePortImpedance', impedance)

    def set_current(self, current: float):
        self.invoke_method('Current', current)

    def set_current_port_admittance(self, admittance: float):
        self.invoke_method('CurrentPortAdmittance', admittance)

    def set_radius(self, radius: float):
        self.invoke_method('Radius', radius)

    def set_point1(self, x: float, y: float, z: float):
        self.invoke_method('SetP1', False, x, y, z)

    def set_pick_as_point1(self):
        self.invoke_method('SetP1', True, 0, 0, 0)

    def set_point2(self, x: float, y: float, z: float):
        self.invoke_method('SetP2', False, x, y, z)

    def set_pick_as_point2(self):
        self.invoke_method('SetP2', True, 0, 0, 0)

    def set_invert_direction(self, flag: bool = True):
        self.invoke_method('InvertDirection', flag)

    def set_local_coordinates(self, flag: bool = True):
        self.invoke_method('LocalCoordinates', flag)

    def set_monitor(self, flag: bool = True):
        self.invoke_method('Monitor', flag)

    def set_wire(self, wire_name: str):
        self.invoke_method('Wire', wire_name)

    def set_wire_end_to_end1(self):
        self.invoke_method('Position', 'end1')

    def set_wire_end_to_end2(self):
        self.invoke_method('Position', 'end2')

    def set_lumped_element_name(self, name: str):
        self.invoke_method('LumpedElementName', name)

    def set_delete_lumped_element(self, flag: bool = True):
        self.invoke_method('DeleteLumpedElement', flag)

    def get_element_dir_index(self, port_number: int):
        dir = w32com.create_ref_long()
        index = w32com.create_ref_long()
        self.invoke_method('GetElementDirIndex', port_number, dir, index)
        return (dir.value, index.value)

    def get_element_second_index(self, port_number: int):
        index = w32com.create_ref_long()
        self.invoke_method('GetElement2ndIndex', port_number, index)
        return index.value

    def get_length(self, port_number: int) -> float:
        return self.invoke_method('GetLength', port_number)

    def get_grid_length(self, port_number: int) -> float:
        return self.invoke_method('GetGridLength', port_number)

    def get_coordinates(self, port_number: int):
        x0 = w32com.create_ref_double()
        y0 = w32com.create_ref_double()
        z0 = w32com.create_ref_double()
        x1 = w32com.create_ref_double()
        y1 = w32com.create_ref_double()
        z1 = w32com.create_ref_double()
        success = self.invoke_method('GetCoordinates', port_number, x0, y0, z0, x1, y1, z1)
        return (x0.value, y0.value, z0.value, x1.value, y1.value, z1.value) if success else None

    def get_properties(self, port_number: int):
        port_type = w32com.create_ref_str()
        impedance = w32com.create_ref_double()
        current = w32com.create_ref_double()
        voltage = w32com.create_ref_double()
        voltage_port_impedance = w32com.create_ref_double()
        radius = w32com.create_ref_double()
        monitor = w32com.create_ref_bool()
        success = self.invoke_method(
            'GetProperties', port_number, port_type, impedance, current, voltage,
            voltage_port_impedance, radius, monitor)
        return (
            port_type.value, impedance.value, current.value, voltage.value,
            voltage_port_impedance.value, radius.value, monitor.value) if success else None

    def delete(self, port_number: int):
        self.invoke_method('Delete', port_number)


from . import Project
from . import ComObjectWrapper
import numpy as np

class Rotate(ComObjectWrapper):
    MODE_POINT_LIST = 'pointlist'
    MODE_PICKS = 'picks'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Rotate

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_component(self, component_name: str):
        self.invoke_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.invoke_method('Material', mat_name)

    def set_mode(self, mode: str):
        self.invoke_method('Mode', mode)

    def set_start_angle_deg(self, angle_deg: float):
        self.invoke_method('StartAngle', angle_deg)

    def set_start_angle_rad(self, angle_rad: float):
        self.set_start_angle_deg(np.rad2deg(angle_rad))

    def set_angle_deg(self, angle_deg: float):
        self.invoke_method('Angle', angle_deg)

    def set_angle_rad(self, angle_rad: float):
        self.set_angle_deg(np.rad2deg(angle_rad))

    def set_height(self, height: float):
        self.invoke_method('Height', height)

    def set_origin(self, x: float, y: float, z: float):
        self.invoke_method('Origin', x, y, z)

    def set_r_vector(self, u: float, v: float, w: float):
        self.invoke_method('Rvector', u, v, w)

    def set_z_vector(self, u: float, v: float, w: float):
        self.invoke_method('Zvector', u, v, w)

    def set_first_point(self, u: float, v: float):
        self.invoke_method('Point', u, v)

    def add_line_to(self, u: float, v: float):
        self.invoke_method('LineTo', u, v)

    def add_line_relative(self, u: float, v: float):
        self.invoke_method('RLine', u, v)

    def set_radius_ratio(self, ratio: float):
        self.invoke_method('RadiusRatio', ratio)

    def modify_angle(self):
        self.invoke_method('ModifyAngle')

    def set_num_steps(self, count: int):
        self.invoke_method('NSteps', count)

    def set_split_closed_edges(self, flag: bool = True):
        self.invoke_method('SplitClosedEdges', flag)

    def set_segmented_profile(self, flag: bool = True):
        self.invoke_method('SegmentedProfile', flag)

    def set_delete_base_face_solid(self, flag: bool = True):
        self.invoke_method('DeleteBaseFaceSolid', flag)

    def set_clear_picked_face(self, flag: bool = True):
        self.invoke_method('ClearPickedFace', flag)

    def set_simplify_solid(self, flag: bool = True):
        self.invoke_method('SimplifySolid', flag)

    def set_use_advanced_segmented_rotation(self, flag: bool = True):
        self.invoke_method('UseAdvancedSegmentedRotation', flag)

    def set_simplify_active(self, flag: bool = True):
        self.invoke_method('SetSimplifyActive', flag)

    def set_simplify_min_points_arc(self, count: int):
        self.invoke_method('SetSimplifyMinPointsArc', count)

    def set_simplify_min_points_circle(self, count: int):
        self.invoke_method('SetSimplifyMinPointsCircle', count)

    def set_simplify_angle_deg(self, angle_deg: float):
        self.invoke_method('SetSimplifyAngle', angle_deg)

    def set_simplify_angle_rad(self, angle_rad: float):
        self.set_simplify_angle_deg(np.rad2deg(angle_rad))

    def set_simplify_adjacent_tol_deg(self, angle_deg: float):
        self.invoke_method('SetSimplifyAdjacentTol', angle_deg)

    def set_simplify_adjacent_tol_rad(self, angle_rad: float):
        self.set_simplify_adjacent_tol_deg(np.rad2deg(angle_rad))

    def set_simplify_radius_tol(self, deviation: float):
        self.invoke_method('SetSimplifyRadiusTol', deviation)

    def set_simplify_angle_tang_deg(self, angle_deg: float):
        self.invoke_method('SetSimplifyAngleTang', angle_deg)

    def set_simplify_angle_tang_rad(self, angle_rad: float):
        self.set_simplify_angle_tang_deg(np.rad2deg(angle_rad))

    def set_simplify_edge_length(self, length: float):
        self.invoke_method('SetSimplifyEdgeLength', length)

    def create(self):
        self.invoke_method('Create')


from . import IVBAProvider, VBAObjWrapper, VBATypeName
from typing import Optional

class Curve(VBAObjWrapper):
    ITERATION_ALL = 'all'
    ITERATION_OPEN = 'open'
    ITERATION_CLOSED = 'closed'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Curve')

    def create_curve(self, name: str):
        self.record_method('NewCurve', name)

    def delete_curve(self, name: str):
        self.record_method('DeleteCurve', name)

    def rename_curve(self, old_name: str, new_name: str):
        self.record_method('RenameCurve', old_name, new_name)

    def delete_curve_item(self, curve_name: str, item_name: str):
        self.record_method('DeleteCurveItem', curve_name, item_name)

    def rename_curve_item(self, curve_name: str, old_item_name: str, new_item_name: str):
        self.record_method('RenameCurveItem', curve_name, old_item_name, new_item_name)

    def delete_curve_item_segment(self, curve_name: str, edge_id: int):
        self.record_method('DeleteCurveItemSegment', curve_name, edge_id)

    def move_curve_item(self, item_name: str, old_curve_name: str, new_curve_name: str):
        self.record_method('MoveCurveItem', item_name, old_curve_name, new_curve_name)

    def start_curve_name_iteration(self, it_type: str):
        self.record_method('StartCurveNameIteration', it_type)

    def get_next_curve_name(self) -> str:
        return self.query_method_str('GetNextCurveName')

    # returns: (x, y, z) or None
    def get_point_coordinates(self, item_name: str, pid: str) \
            -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetPointCoordinates', B, item_name, pid, D, D, D)
        return None if not res[0] else res[1:]

    def get_number_of_points(self) -> int:
        return self.query_method_int('GetNumberOfPoints')

    def is_closed(self, item_name: str) -> bool:
        return self.query_method_bool('IsClosed', item_name)



from . import IVBAProvider, VBAObjWrapper

class FieldSource(VBAObjWrapper):
    ALIGNMENT_CUSTOM = 'user'
    ALIGNMENT_CURRENT_WCS = 'currentwcs'
    ALIGNMENT_SOURCE_FILE = 'sourcefile'

    CALC_MODE_AUTOMATIC = 'automatic'
    CALC_MODE_CUSTOM = 'user defined'
    CALC_MODE_AUTOTRUNCATION = 'autotruncation'

    FIELD_TYPE_TEMPERATURE = 'Temperature'
    FIELD_TYPE_DISPLACEMENT = 'Displacement'
    FIELD_TYPE_FORCE_DENSITY = 'Force Density'
    FIELD_TYPE_NODAL_FORCES = 'Nodal Forces'
    FIELD_TYPE_THERMAL_LOSSES = 'Thermal Losses'

    FIELD_MONITOR_NAME_INITIAL_SOLUTION = 'Initial Solution'
    FIELD_MONITOR_NAME_STATIONARY_SOLUTION = 'Stationary Solution'

    IMPORT_SETTING_FORMAT_SOURCE = 'FormatSource'
    IMPORT_SETTING_FORMAT_TARGET = 'FormatTarget'
    IMPORT_SETTING_NASTRAN_FILE_NAME = 'NastranFilename'
    IMPORT_SETTING_TEMPERATURE_FILE_NAME = 'TemperatureFilename'
    IMPORT_SETTING_GEOMETRY_UNIT = 'GeometryUnit'
    IMPORT_SETTING_TEMPERATURE_UNIT = 'TemperatureUnit'
    IMPORT_SETTING_WORKING_DIR = 'WorkingDir'

    IMPORT_SETTING_FORMAT_SOURCE_ABAQUS = 'Abaqus'
    IMPORT_SETTING_FORMAT_SOURCE_TET = 'Tet'
    IMPORT_SETTING_FORMAT_SOURCE_HEX = 'Hex'
    IMPORT_SETTING_FORMAT_SOURCE_CHT = 'CHT'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'FieldSource')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def set_file_name(self, name: str):
        self.record_method('FileName', name)

    def delete(self, name: str):
        self.record_method('Delete', name)

    def set_id(self, id: int):
        self.record_method('Id', id)

    def get_next_id(self) -> int:
        return self.query_method_int('GetNextId')

    def delete_all(self):
        self.record_method('DeleteAll')

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def read_from_file(self):
        self.record_method('Read')

    def set_project_path(self, path: str):
        self.record_method('ProjectPath', path)

    def set_use_relative_path(self, flag: bool = True):
        self.record_method('UseRelativePath', flag)

    def set_result_subdir(self, path: str):
        self.record_method('ResultSubDirectory', path)

    def set_field_type(self, field_type: str):
        self.record_method('SourceName', field_type)

    def set_field_monitor_name(self, name: str):
        self.record_method('FieldMonitorName', name)

    def set_import_setting(self, key: str, value: str):
        self.record_method('ImportSettings', key, value)

    def set_time_value(self, time: float):
        self.record_method('TimeValue', time)

    def set_use_last_time_frame(self, flag: bool = True):
        self.record_method('UseLastTimeFrame', flag)

    def set_use_copy_only(self, flag: bool = True):
        self.record_method('UseCopyOnly', flag)

    def create_field_import(self):
        self.record_method('CreateFieldImport')

    def create_field_import_from_abaqus(
            self, name: str, dir: str, nastran_file_name: str, distribution_file_name: str,
            geom_unit: str, temp_unit: str):
        self.record_method(
            'CreateFieldImportFromAbaqus', name, dir, nastran_file_name, distribution_file_name,
            geom_unit, temp_unit)


from . import Project
from . import ComObjectWrapper

class Loft(ComObjectWrapper):
    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Loft

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_name(self, name: str):
        self.invoke_method('Name', name)

    def set_component(self, component_name: str):
        self.invoke_method('Component', component_name)

    def set_material(self, mat_name: str):
        self.invoke_method('Material', mat_name)

    def set_tangency(self, tang: float):
        self.invoke_method('Tangency', tang)

    def create(self):
        #TODO: check if Create or CreateNew
        self.invoke_method('Create')



from . import IVBAProvider, VBAObjWrapper

class Arc(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Arc')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, arc_name: str):
        self.record_method('Name', arc_name)

    def set_curve(self, curve_name: str):
        self.record_method('Curve', curve_name)

    def set_orientation_clockwise(self):
        self.record_method('Orientation', 'Clockwise')

    def set_orientation_counter_clockwise(self):
        self.record_method('Orientation', 'CounterClockwise')

    def set_center_x(self, x: float):
        self.record_method('Xcenter', x)

    def set_center_y(self, y: float):
        self.record_method('Ycenter', y)

    def set_center(self, x: float, y: float):
        self.set_center_x(x)
        self.set_center_y(y)

    def set_start_point_x(self, x: float):
        self.record_method('X1', x)

    def set_start_point_y(self, y: float):
        self.record_method('Y1', y)

    def set_start_point_x(self, x: float, y: float):
        self.set_start_point_x(x)
        self.set_start_point_y(y)


from . import Project
from . import ComObjectWrapper

class IESolver(ComObjectWrapper):
    ACCURACY_CUSTOM = 'Custom'
    ACCURACY_LOW = 'Low'
    ACCURACY_MEDIUM = 'Medium'
    ACCURACY_HIGH = 'High'

    REAL_GROUND_MODEL_AUTO = 'Auto'
    REAL_GROUND_MODEL_TYPE1 = 'Type 1'
    REAL_GROUND_MODEL_TYPE2 = 'Type 2'

    PRECONDITIONER_AUTO = 'Auto'
    PRECONDITIONER_TYPE1 = 'Type 1'
    PRECONDITIONER_TYPE2 = 'Type 2'
    PRECONDITIONER_TYPE3 = 'Type 3'

    CMA_ACCURACY_DEFAULT = 'Default'
    CMA_ACCURACY_CUSTOM = 'Custom'

    CMA_MEM_CUSTOM = 'Custom'
    CMA_MEM_LOW = 'Low'
    CMA_MEM_MEDIUM = 'Medium'
    CMA_MEM_HIGH = 'High'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.IESolver

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def reset(self):
        self.invoke_method('Reset')

    def set_accuracy_setting(self, accuracy: str):
        self.invoke_method('SetAccuracySetting', accuracy)

    def set_use_fast_frequency_sweep(self, flag: bool = True):
        self.invoke_method('UseFastFrequencySweep', flag)

    def set_use_ie_ground_plane(self, flag: bool = True):
        self.invoke_method('UseIEGroundPlane', flag)

    def set_real_ground_material_name(self, name: str):
        if name is None:
            name = ''
        self.invoke_method('SetRealGroundMaterialName', name)

    def set_calc_farfield_in_real_ground(self, flag: bool = True):
        self.invoke_method('CalcFarFieldInRealGround', flag)

    def set_real_ground_model_type(self, model_type: str):
        self.invoke_method('RealGroundModelType', model_type)

    def set_preconditioner_type(self, preconditioner_type: str):
        self.invoke_method('PreconditionerType', preconditioner_type)

    def set_low_frequency_stabilization(self, flag: bool = True):
        self.invoke_method('LowFrequencyStabilization', flag)

    def set_low_frequency_stabilization_multilayer(self, flag: bool = True):
        self.invoke_method('LowFrequencyStabilizationML', flag)

    def set_multilayer(self, flag: bool = True):
        self.invoke_method('Multilayer', flag)

    def set_iterative_mom_accuracy_ie_solver(self, accuracy: float):
        self.invoke_method('SetiMoMACC_I', accuracy)

    def set_iterative_mom_accuracy_multilayer_solver(self, accuracy: float):
        self.invoke_method('SetiMoMACC_M', accuracy)

    def set_cfie_alpha(self, accuracy: float):
        self.invoke_method('SetCFIEAlpha', accuracy)

    def set_deembed_external_ports(self, flag: bool = True):
        self.invoke_method('DeembedExternalPorts', flag)

    def set_open_boundary_condition_in_xy_dir(self, flag: bool = True):
        self.invoke_method('SetOpenBC_XY', flag)

    def set_cma_mode_tracking(self, flag: bool = True):
        self.invoke_method('ModeTrackingCMA', flag)

    def set_cma_number_of_modes(self, number: int):
        self.invoke_method('NumberOfModesCMA', number)

    def set_cma_start_frequency(self, freq: float):
        self.invoke_method('StartFrequencyCMA', freq)

    def set_cma_accuracy_setting(self, accuracy: str):
        self.invoke_method('SetAccuracySettingCMA', accuracy)

    def set_cma_number_of_frequency_samples(self, number: int):
        self.invoke_method('FrequencySamplesCMA', number)

    def set_cma_mem_setting(self, mem: str):
        self.invoke_method('SetMemSettingCMA', mem)

    def set_cma_calculate_modal_weighting_coefficients(self, flag: bool = True):
        self.invoke_method('CalculateModalWeightingCoefficientsCMA', flag)


from . import IVBAProvider, VBATypeName
from typing import Union, Optional

class Parameters:
    def __init__(self, vbap: IVBAProvider) -> None:
        self.vbap = vbap

    def invoke_fn(self, name: str, *args) -> None:
        self.vbap.invoke_function(name, *args)

    def query_fn(self, name: str, *args) -> str:
        self.vbap.query_function(name, *args)

    def delete(self, name: str):
        self.invoke_fn('DeleteParameter', name)

    def has(self, name: str) -> bool:
        return self.query_fn('DoesParameterExist', name) != 'False'

    def get_count(self) -> int:
        return int(self.query_fn('GetNumberOfParameters'))

    def get_name(self, index: int) -> str:
        return self.query_fn('GetParameterName', index)

    def get_value(self, index: int) -> float:
        return float(self.query_fn('GetParameterNValue', index))

    def get_expression(self, index: int) -> str:
        return self.query_fn('GetParameterSValue', index)

    def rename(self, old_name: str, new_name: str):
        self.invoke_fn('RenameParameter', old_name, new_name)

    def restore(self, name: str) -> str:
        return self.query_fn('RestoreParameter', name)

    def restore_as_number(self, name: str) -> float:
        return float(self.query_fn('RestoreDoubleParameter', name))

    def restore_expression(self, name: str) -> str:
        return self.query_fn('RestoreParameterExpression', name)

    def store(
            self, name: str, expression: Union[str, int, float],
            description: Optional[str] = None):
        if description is None:
            self.invoke_fn('StoreParameter', name, str(expression))
        else:
            self.invoke_fn('StoreParameterWithDescription', name, expression, description)

    def ensure_exists(self, name: str, default_expression: Union[str, int, float]):
        self.invoke_fn('MakeSureParameterExists', name, default_expression)

    def does_project_depend_on(self, param_name: str) -> bool:
        return self.query_fn('DoesProjectDependOnParameter', param_name) != 'False'

    def set_description(self, name: str, description: str):
        self.invoke_fn('SetParameterDescription', name, description)

    def get_description(self, name: str) -> str:
        return self.query_fn('GetParameterDescription', name)

    # returns: (names_var, values_var) or None
    def get_combination(self, result_id: str) -> Optional[tuple[str, str]]:
        B, V = VBATypeName.Boolean, VBATypeName.Variant
        res = self.vbap.query_function_t('GetParameterCombination', B, result_id, V, V)
        return None if not res[0] else res[1:]


from py4cst.CST import Material
from py4cst.CST import units
from typing import Optional

def set_normal_type_general_properties(
        material: Material,
        rel_permitivity: float, rel_permeability: float,
        name: Optional[str] = None, folder: Optional[str] = None):
    material.set_type(Material.TYPE_NORMAL)
    material.set_rel_permitivity(rel_permitivity)
    material.set_rel_permeability(rel_permeability)
    if name is not None:
        material.set_name(name)
    if folder is not None:
        material.set_folder_name(folder)

def set_appearance(
        material: Material,
        color_rgba: tuple[float, float, float, float] = (0.6, 0.6, 0.6, 1.0),
        wireframe: bool = False, allow_outlines: bool = True,
        reflection: bool = False, transparent_solid_outlines: bool = False):
    material.set_color_rgba(*color_rgba)
    material.set_wireframe(wireframe)
    material.set_allow_outlines(allow_outlines)
    material.set_reflection(reflection)
    material.set_transparent_solid_outlines(transparent_solid_outlines)

def set_units(
        material: Material,
        freq_unit: str = units.FREQUENCY_HERTZ, geom_unit: str = units.GEOMETRY_METER,
        time_unit: str = units.TIME_SECOND, temp_unit: str = units.TEMPERATURE_KELVIN):
    material.set_frequency_unit(freq_unit)
    material.set_geometry_unit(geom_unit)
    material.set_time_unit(time_unit)
    material.set_temperature_unit(temp_unit)

def set_simple_el_conductivity(material: Material, conductivity: float):
    material.set_el_conductivity(conductivity)
    material.set_el_tangent_delta_given(False)
    material.set_el_parametric_conductivity(False)

def set_simple_mag_conductivity(material: Material, conductivity: float):
    material.set_mag_conductivity(conductivity)
    material.set_mag_tangent_delta_given(False)
    material.set_mag_parametric_conductivity(False)

def set_const_el_tangent_delta(material: Material, tand: float, tand_freq: float):
    material.set_el_tangent_delta(tand)
    material.set_el_tangent_delta_frequency(tand_freq)
    material.set_el_tangent_delta_given(True)
    material.set_el_tangent_delta_model(Material.TANGENT_DELTA_MODEL_CONST_TAN_D)
    material.set_el_const_tangent_delta_strategy_eps(
        Material.CONST_TANGENT_DELTA_STRATEGY_AUTOMATIC_ORDER)

def set_const_mag_tangent_delta(material: Material, tand: float, tand_freq: float):
    material.set_mag_tangent_delta(tand)
    material.set_mag_tangent_delta_frequency(tand_freq)
    material.set_mag_tangent_delta_given(True)
    material.set_mag_tangent_delta_model(Material.TANGENT_DELTA_MODEL_CONST_TAN_D)
    material.set_mag_const_tangent_delta_strategy_mu(
        Material.CONST_TANGENT_DELTA_STRATEGY_AUTOMATIC_ORDER)

def prepare_simple_material(
        material: Material,
        rel_permitivity: float = 1.0, rel_permeability: float = 1.0,
        el_conductivity: float = 0.0, mag_conductivity: float = 0.0) -> None:
    material.reset()
    material.set_material_density(0)
    material.set_reference_coord_system(Material.REF_COORD_SYSTEM_GLOBAL)
    material.set_coord_system_type(Material.COORD_SYSTEM_CARTESIAN)
    set_units(material)
    set_normal_type_general_properties(material, rel_permitivity, rel_permeability)
    set_simple_el_conductivity(material, el_conductivity)
    set_simple_mag_conductivity(material, mag_conductivity)

def prepare_simple_material_tand(
        material: Material,
        rel_permitivity: float = 1.0, rel_permeability: float = 1.0,
        el_tan_delta: float = 0.0, mag_tan_delta: float = 0.0,
        el_tan_delta_freq: float = 0.0, mag_tan_delta_freq: float = 0.0) -> None:
    material.reset()
    material.set_material_density(0)
    material.set_reference_coord_system(Material.REF_COORD_SYSTEM_GLOBAL)
    material.set_coord_system_type(Material.COORD_SYSTEM_CARTESIAN)
    set_units(material)
    set_normal_type_general_properties(material, rel_permitivity, rel_permeability)
    set_const_el_tangent_delta(material, el_tan_delta, el_tan_delta_freq)
    set_const_mag_tangent_delta(material, mag_tan_delta, mag_tan_delta_freq)

def prepare_vacuum(material: Material) -> None:
    prepare_simple_material(
        material,
        rel_permitivity=1.0, rel_permeability=1.0,
        el_conductivity=0.0, mag_conductivity=0.0)

def prepare_surface_impedance_table(
        material: Material,
        frq_imp_weight: list[tuple[float, complex, float]],
        max_order: int = 10,
        error_limit: float = 0.1,
        transparent_model: bool = False):
    # frq_imp_weight: list of tuples: [(freq, impdance, weight), ...]
    material.reset()
    material.set_material_density(0)
    material.set_reference_coord_system(Material.REF_COORD_SYSTEM_GLOBAL)
    material.set_coord_system_type(Material.COORD_SYSTEM_CARTESIAN)
    set_units(material)
    material.set_type(Material.TYPE_LOSSY_METAL)
    material.set_tabulated_surface_impedance_model(
        Material.TAB_SURF_IMP_MODEL_TRANSPARENT\
            if transparent_model else Material.TAB_SURF_IMP_MODEL_OPAQUE)
    for freq, imp, weight in frq_imp_weight:
        material.add_tabulated_surface_impedance_fitting_value(freq, imp, weight)
    material.set_dispersive_fitting_scheme_tab_surf_imp(Material.DISP_FIT_SCHEME_TAB_SI_NTH_ORDER)
    material.set_max_order_nth_model_fit_tab_surf_imp(max_order)
    material.set_use_only_data_in_sim_freq_range_nth_model_tab_surf_imp(True)
    material.set_error_limit_nth_model_fit_tab_surf_imp(error_limit)

# With Material
#      .Reset
#      .Name "material1"
#      .Folder ""
#      .Rho "0.0"
#      .ThermalType "Normal"
#      .ThermalConductivity "0"
#      .SpecificHeat "0", "J/K/kg"
#      .DynamicViscosity "0"
#      .UseEmissivity "True"
#      .Emissivity "0"
#      .MetabolicRate "0.0"
#      .VoxelConvection "0.0"
#      .BloodFlow "0"
#      .MechanicsType "Unused"
#      .IntrinsicCarrierDensity "0"
#      .FrqType "all"
#      .Type "Lossy metal"
#      .MaterialUnit "Frequency", "GHz"
#      .MaterialUnit "Geometry", "mm"
#      .MaterialUnit "Time", "ns"
#      .MaterialUnit "Temperature", "Celsius"
#      .SetTabulatedSurfaceImpedanceModel "Opaque"
#      .AddTabulatedSurfaceImpedanceFittingValue "10", "1.1", "-3.2", "1.0"
#      .ReferenceCoordSystem "Global"
#      .CoordSystemType "Cartesian"
#      .DispersiveFittingSchemeTabSI "Nth Order"
#      .MaximalOrderNthModelFitTabSI "9"
#      .ErrorLimitNthModelFitTabSI "0.1"
#      .UseOnlyDataInSimFreqRangeNthModelTabSI "False"
#      .NLAnisotropy "False"
#      .NLAStackingFactor "1"
#      .NLADirectionX "1"
#      .NLADirectionY "0"
#      .NLADirectionZ "0"
#      .Colour "0", "1", "1"
#      .Wireframe "False"
#      .Reflection "False"
#      .Allowoutline "True"
#      .Transparentoutline "False"
#      .Transparency "0"
#      .Create
# End With

# def prepare_simple_material(
#         material: Material,
#         rel_permitivity: float = 1.0, rel_permeability: float = 1.0,
#         el_conductivity: float = 0.0, mag_conductivity: float = 0.0) -> None:
#     material.reset()
#     material.set_material_density(0)
#     material.set_thermal_type(Material.THERMAL_TYPE_NORMAL)
#     material.set_thermal_conductivity(0)
#     material.set_specific_heat(0, Material.SPECIFIC_HEAT_UNIT_J_K_KG)
#     material.set_dynamic_viscosity(0)
#     material.set_use_emissivity(True)
#     material.set_emissivity(0)
#     material.set_metabolic_rate(0)
#     material.set_voxel_convection(0)
#     material.set_blood_flow(0)
#     material.set_mechanics_type(Material.MECHANICS_TYPE_UNUSED)
#     material.set_intrinsic_carrier_density(0)
#     material.set_reference_coord_system(Material.REF_COORD_SYSTEM_GLOBAL)
#     material.set_coord_system_type(Material.COORD_SYSTEM_CARTESIAN)
#     set_units(material)
#     set_normal_type_general_properties(material, rel_permitivity, rel_permeability)
#     set_simple_el_conductivity(material, el_conductivity)
#     set_simple_mag_conductivity(mag_conductivity)
#     material.set_disp_model_esp(Material.DISP_MODEL_EPS_NONE)
#     material.set_disp_model_mu(Material.DISP_MODEL_MU_NONE)
#     material.set_dispersive_fitting_scheme_eps(Material.DISPERSIVE_FITTING_SCHEME_NTH_ORDER)
#     material.set_max_order_nth_model_fit_eps(10)
#     material.set_error_limit_nth_model_fit_esp(0.1)
#     material.set_use_only_data_in_sim_freq_range_nth_model_eps(False)
#     material.set_dispersive_fitting_scheme_mu(Material.DISPERSIVE_FITTING_SCHEME_NTH_ORDER)
#     material.set_max_order_nth_model_fit_mu(10)
#     material.set_error_limit_nth_model_fit_mu(0.1)
#     material.set_use_only_data_in_sim_freq_range_nth_model_mu(False)
#     material.set_use_general_dispersion_eps(False)
#     material.set_use_general_dispersion_mu(False)
#     material.set_use_nl_anisotropy(False)
#     material.set_nla_stacking_factor(1)
#     material.set_nla_direction(1, 0, 0)
#     set_appearance(material)


from . import IVBAProvider, VBAObjWrapper

class LayerStacking(VBAObjWrapper):
    MATERIAL_NORMAL = 'normal'
    MATERIAL_PEC = 'pec'

    THERMAL_TYPE_NORMAL = 'normal'
    THERMAL_TYPE_PTC = 'ptc'

    NORMAL_DIRECTION_X = 'x'
    NORMAL_DIRECTION_Y = 'y'
    NORMAL_DIRECTION_Z = 'z'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'LayerStacking')

    def reset(self):
        self.record_method('Reset')

    def set_layer_stacking_active(self, flag: bool = True):
        self.record_method('LayerStackingActive', flag)

    def set_background_items_align_value(self, value: float):
        self.record_method('AlignValueBackgroundItems', value)

    def set_background_items_normal(self, direction: str):
        self.record_method('NormalBackgroundItems', direction)

    def set_invert_direction(self, flag: bool = True):
        self.record_method('InvertDirection', flag)

    def set_fix_traversal(self, flag: bool = True):
        self.record_method('FixTransversal', flag)

    def add_item(self, index: int, height: float, material_name: str):
        self.record_method('AddItem', index, height, material_name)


from . import IHistoryListProvider
from typing import Optional, Union

class HistoryItem:
    def __init__(
            self, object_name: Optional[str], method_name: str,
            *method_args: Union[str, int, float, bool]) -> None:
        self.object_name = object_name
        self.method_name = method_name
        self.method_args = list(method_args)

    def to_vba_code(self) -> str:
        obj_name = '' if self.object_name is None else f'{self.object_name}.'
        args = [f'"{arg}"' if isinstance(arg, str) else str(arg) for arg in self.method_args]
        args_str = ", ".join(args)
        return f'{obj_name}{self.method_name}({args_str})'

class HistoryListGenerator:
    def __init__(self, history_list_provider: IHistoryListProvider) -> None:
        self.names = {}
        self.history: list[HistoryItem] = []
        self.history_list_provider = history_list_provider

    def invoke_method(self, object_name, method_name, *args) -> None:
        self.generate_item(HistoryItem(object_name, method_name, *args))

    def cache_method(self, object_name, method_name, *args) -> None:
        self.history.append(HistoryItem(object_name, method_name, *args))

    def generate_cache(self):
        for item in self.history:
            self.generate_item(item)

    def generate_item(self, item: HistoryItem):
        name = '' if item.object_name is None else f'{item.object_name}.'
        name += item.method_name
        id = self.names.get(name, 0)
        self.names[name] = id + 1
        name += f':{id}'
        self.history_list_provider.add_to_history(name, item.to_vba_code())

class IHistoryListGeneratorProvider:
    def get_history_list_generator(self) -> HistoryListGenerator:
        pass


from . import HistoryListGenerator, VBATypeName, IQuietModeController
from typing import Optional, Union

class IVBAProvider:
    def get_history_list_generator(self) -> HistoryListGenerator:
        pass

    def run_vba(self, vba_code: str) -> bool:
        pass

    def invoke_function(self, function_name: str, *args: Union[str, int, float, bool]) -> bool:
        pass

    def invoke_method(
            self, object_name: str, function_name: str, *args: Union[str, int, float, bool]):
        pass

    def query_vba(self, vba_code: str) -> Optional[str]:
        pass

    def query_function(
            self, function_name: str, *args: Union[str, int, float, bool]) -> Optional[str]:
        pass

    def query_method(
            self, object_name: str, function_name: str,
            *args: Union[str, int, float, bool]) -> Optional[str]:
        pass

    def query_function_t(
            self, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]) -> tuple:
        pass

    def query_method_t(
            self, object_name: str, function_name: str, return_type: Optional[VBATypeName],
            *args: Union[str, int, float, bool, VBATypeName]) -> tuple:
        pass

    def get_quiet_mode_controller(self) -> IQuietModeController:
        pass


from . import IVBAProvider, VBAObjWrapper

class AnalyticalCurve(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'AnalyticalCurve')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, analytical_curve_name: str):
        self.record_method('Name', analytical_curve_name)

    def set_curve(self, curve_name: str):
        self.record_method('Curve', curve_name)

    def set_law_x(self, law: str):
        self.record_method('LawX', law)

    def set_law_y(self, law: str):
        self.record_method('LawY', law)

    def set_law_z(self, law: str):
        self.record_method('LawZ', law)

    def set_parameter_range(self, min: float, max: float):
        self.record_method('ParameterRange', min, max)

    def create(self):
        self.record_method('Create')


from . import IVBAProvider, VBAObjWrapper, VBATypeName
import numpy as np
from typing import Optional

class WCS(VBAObjWrapper):
    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'WCS')

    def activate_local(self):
        self.record_method('ActivateWCS', 'local')

    def activate_global(self):
        self.record_method('ActivateWCS', 'global')

    def store(self, name: str):
        self.record_method('Store', name)

    def restore(self, name: str):
        self.record_method('Restore', name)

    def delete(self, name: str):
        self.record_method('Delete', name)

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def set_normal(self, x: float, y: float, z: float):
        self.record_method('SetNormal', x, y, z)

    def set_origin(self, x: float, y: float, z: float):
        self.record_method('SetOrigin', x, y, z)

    def set_u_vector(self, x: float, y: float, z: float):
        self.record_method('SetUVector', x, y, z)

    def align_wcs_with_selected_point(self):
        self.record_method('AlignWCSWithSelected', 'Point')

    def align_wcs_with_selected_3_points(self):
        self.record_method('AlignWCSWithSelected', '3Points')

    def align_wcs_with_selected_edge(self):
        self.record_method('AlignWCSWithSelected', 'Edge')

    def align_wcs_with_selected_edge_center(self):
        self.record_method('AlignWCSWithSelected', 'EdgeCenter')

    def align_wcs_with_selected_rotation_edge(self):
        self.record_method('AlignWCSWithSelected', 'RotationEdge')

    def align_wcs_with_selected_face(self):
        self.record_method('AlignWCSWithSelected', 'Face')

    def rotate_wcs_around_u_deg(self, angle_deg: float):
        self.record_method('RotateWCS', 'u', angle_deg)

    def rotate_wcs_around_u_rad(self, angle_rad: float):
        self.rotate_wcs_around_u_deg(np.rad2deg(angle_rad))

    def rotate_wcs_around_v_deg(self, angle_deg: float):
        self.record_method('RotateWCS', 'v', angle_deg)

    def rotate_wcs_around_v_rad(self, angle_rad: float):
        self.rotate_wcs_around_v_deg(np.rad2deg(angle_rad))

    def rotate_wcs_around_w_deg(self, angle_deg: float):
        self.record_method('RotateWCS', 'w', angle_deg)

    def rotate_wcs_around_w_rad(self, angle_rad: float):
        self.rotate_wcs_around_w_deg(np.rad2deg(angle_rad))

    def move_local_wcs(self, du: float, dv: float, dw: float):
        self.record_method('MoveWCS', 'local', du, dv, dw)

    def move_global_wcs(self, du: float, dv: float, dw: float):
        self.record_method('MoveWCS', 'global', du, dv, dw)

    def align_wcs_with_global_coordinates(self):
        self.record_method('AlignWCSWithGlobalCoordinates')

    def set_workplane_size(self, size: float):
        self.record_method('SetWorkplaneSize', size)

    def set_workplane_raster(self, raster_size: float):
        self.record_method('SetWorkplaneRaster', raster_size)

    def set_workplane_snap(self, flag: bool = True):
        self.record_method('SetWorkplaneSnap', flag)

    def set_workplane_autoadjust(self, flag: bool = True):
        self.record_method('SetWorkplaneAutoadjust', flag)

    def set_workplane_snap_autoadjust(self, flag: bool = True):
        self.record_method('SetWorkplaneSnapAutoadjust', flag)

    def set_workplane_autosnap_factor(self, factor: float):
        self.record_method('SetWorkplaneAutosnapFactor', factor)

    def set_workplane_snap_raster(self, raster_size: float):
        self.record_method('SetWorkplaneSnapRaster', raster_size)

    def is_local_wcs_active(self) -> bool:
        return self.query_method_str('IsWCSActive') == 'local'

    def is_global_wcs_active(self) -> bool:
        return self.query_method_str('IsWCSActive') == 'global'

    def does_exist(self, wcs_name: str) -> bool:
        return self.query_method_bool('DoesExist', wcs_name)

    # returns: (x, y, z)
    def get_origin(self, wcs_name: str) -> tuple[float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetOrigin', None, wcs_name, D, D, D)

    def get_normal(self, wcs_name: str) -> tuple[float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetNormal', None, wcs_name, D, D, D)

    def get_u_vector(self, wcs_name: str) -> tuple[float, float, float]:
        D = VBATypeName.Double
        return self.query_method_t('GetUVector', None, wcs_name, D, D, D)

    # returns: (ux, uy, yz, vx, vy, vz, wx, wy, wz) or None
    def get_affine_matrix_uvw_to_xyz(self, wcs_name: str) \
            -> Optional[tuple[float, float, float, float, float, float, float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetAffineMatrixUVW2XYZ', B, wcs_name, D, D, D, D, D, D, D, D, D)
        return None if not res[0] else res[1:]

    # returns: (ux, uy, yz, vx, vy, vz, wx, wy, wz) or None
    def get_affine_matrix_xyz_to_uvw(self, wcs_name: str) \
            -> Optional[tuple[float, float, float, float, float, float, float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetAffineMatrixXYZ2UVW', B, wcs_name, D, D, D, D, D, D, D, D, D)
        return None if not res[0] else res[1:]

    # returns (u, v, w) or None
    def get_wcs_point_from_global(self, wcs_name: str, x: float, y: float, z: float) \
            -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetWCSPointFromGlobal', B, wcs_name, D, D, D, x, y, z)
        return None if not res[0] else res[1:]

    # returns: (x, y, z) or None
    def get_global_point_from_wcs(self, wcs_name: str, u: float, v: float, w: float) \
            -> Optional[tuple[float, float, float]]:
        B, D = VBATypeName.Boolean, VBATypeName.Double
        res = self.query_method_t('GetGlobalPointFromWCS', B, wcs_name, D, D, D, u, v, w)
        return None if not res[0] else res[1:]


from pywintypes import com_error

class ComObjectWrapper:
    def invoke_method(self, name, *args, **kwargs):
        try:
            self.com_object._FlagAsMethod(name)
            method = self.invoke_attribute(name)
            return method(*args, **kwargs)
        except com_error as e:
            # raise RuntimeError(e.args[2][2])
            raise RuntimeError(e)

    def invoke_attribute(self, name):
        return getattr(self.com_object, name)


from . import IVBAProvider, VBAObjWrapper

class Transform(VBAObjWrapper):
    TRANSFORM_WHAT_SHAPE = 'Shape'
    TRANSFORM_WHAT_ANCHOR_POINT = 'Anchorpoint'
    TRANSFORM_WHAT_FACE = 'Face'
    TRANSFORM_WHAT_MESH_SHAPE = 'Meshshape'
    TRANSFORM_WHAT_PROBE = 'Probe'
    TRANSFORM_WHAT_VOXEL_DATA = 'Voxeldata'
    TRANSFORM_WHAT_MIXED = 'mixed'
    TRANSFORM_WHAT_FFS = 'FFS'
    TRANSFORM_WHAT_HF_3D_MONITOR = 'HF3DMonitor'
    TRANSFORM_WHAT_PORT = 'Port'
    TRANSFORM_WHAT_LUMPED_ELEMENT = 'Lumpedelement'
    TRANSFORM_WHAT_CURRENT_DISTRIBUTION = 'Currentdistribution'
    TRANSFORM_WHAT_COIL = 'Coil'
    TRANSFORM_WHAT_CURRENT_MONITOR = 'currentmonitor'
    TRANSFORM_WHAT_CURRENT_WIRE = 'currentwire'
    TRANSFORM_WHAT_VOLTAGE_MONITOR = 'voltagemonitor'
    TRANSFORM_WHAT_VOLTAGE_WIRE = 'voltagewire'

    TRANSFORM_HOW_TRANSLATE = 'Translate'
    TRANSFORM_HOW_ROTATE = 'Rotate'
    TRANSFORM_HOW_SCALE = 'Scale'
    TRANSFORM_HOW_MIRROR = 'Mirror'
    TRANSFORM_HOW_MATRIX = 'Matrix'
    TRANSFORM_HOW_LOCAL_TO_GLOBAL = 'LocalToGlobal'
    TRANSFORM_HOW_GLOBAL_TO_LOCAL = 'GlobalToLocal'

    ORIGIN_SHAPE_CENTER = 'ShapeCenter'
    ORIGIN_COMMON_CENTER = 'CommonCenter'
    ORIGIN_FREE = 'Free'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Transform')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def add_name(self, name: str):
        self.record_method('AddName', name)

    def translate_curve(self):
        self.record_method('TranslateCurve')

    def scale_curve(self):
        self.record_method('ScaleCurve')

    def rotate_curve(self):
        self.record_method('RotateCurve')

    def mirror_curve(self):
        self.record_method('MirrorCurve')

    def translate_wire(self):
        self.record_method('TranslateWire')

    def scale_wire(self):
        self.record_method('ScaleWire')

    def rotate_wire(self):
        self.record_method('RotateWire')

    def mirror_wire(self):
        self.record_method('MirrorWire')

    def translate_coil(self):
        self.record_method('TranslateCoil')

    def scale_coil(self):
        self.record_method('ScaleCoil')

    def rotate_coil(self):
        self.record_method('RotateCoil')

    def mirror_coil(self):
        self.record_method('MirrorCoil')

    def transform(self, what: str, how: str):
        self.record_method('Transform', what, how)

    def set_use_picked_points(self, flag: bool = True):
        self.record_method('UsePickedPoints', flag)

    def set_invert_picked_points(self, flag: bool = True):
        self.record_method('InvertPickedPoints', flag)

    def set_multiple_objects(self, flag: bool = True):
        self.record_method('MultipleObjects', flag)

    def set_group_objects(self, flag: bool = True):
        self.record_method('GroupObjects', flag)

    def set_origin(self, key: str):
        self.record_method('Origin', key)

    def set_center(self, u: float, v: float, w: float):
        self.record_method('Center', u, v, w)

    def set_vector(self, u: float, v: float, w: float):
        self.record_method('Vector', u, v, w)

    def set_scale_factor(self, u: float, v: float, w: float):
        self.record_method('ScaleFactor', u, v, w)

    def set_angle(self, u: float, v: float, w: float):
        self.record_method('Angle', u, v, w)

    def set_plane_normal(self, u: float, v: float, w: float):
        self.record_method('PlaneNormal', u, v, w)

    def set_matrix(
            self,
            c11: float, c12: float, c13: float,
            c21: float, c22: float, c23: float,
            c31: float, c32: float, c33: float):
        self.record_method('Matrix', c11, c12, c13, c21, c22, c23, c31, c32, c33)

    def set_number_of_repetitions(self, count: int):
        self.record_method('Repetitions', count)

    def set_component(self, name: str):
        self.record_method('Component', name)

    def set_material(self, name: str):
        self.record_method('Material', name)

    def set_multiple_selection(self, flag: bool = True):
        self.record_method('MultipleSelection', flag)

    def set_destination(self, destination: str):
        self.record_method('Destination', destination)

    def set_auto_destination(self, flag: bool = True):
        #NOTE: officially undocumented
        self.record_method('AutoDestination', flag)

    def set_touch(self, flag: bool = True):
        self.record_method('Touch', flag)

    def add_name_to_active_touch_set(self, name: str):
        self.record_method('AddNameToActiveTouchSet', name)

    def add_name_to_passive_touch_set(self, name: str):
        self.record_method('AddNameToPassiveTouchSet', name)

    def set_touch_tolerance(self, tolerance: float):
        self.record_method('TouchTolerance', tolerance)

    def set_touch_max_num_iterations(self, count: int):
        self.record_method('TouchMaxIterations', count)

    def set_touch_heuristic(self, flag: bool = True):
        self.record_method('TouchHeuristic', flag)

    def set_touch_offset(self, offset: float):
        self.record_method('TouchOffset', offset)


from . import IVBAProvider, VBAObjWrapper
import numpy as np

class Wire(VBAObjWrapper):
    TYPE_BOND_WIRE = 'Bondwire'
    TYPE_CURVE_WIRE = 'Curvewire'

    BOND_WIRE_TYPE_SPLINE = 'Spline'
    BOND_WIRE_TYPE_JEDEC4 = 'JEDEC4'
    BOND_WIRE_TYPE_JEDEC5 = 'JEDEC5'

    TERMINATION_NATURAL = 'natural'
    TERMINATION_ROUNDED = 'rounded'
    TERMINATION_EXTENDED = 'extended'

    def __init__(self, vbap: IVBAProvider) -> None:
        super().__init__(vbap, 'Wire')

    def reset(self):
        self.record_method('Reset')

    def set_name(self, name: str):
        self.record_method('Name', name)

    def set_folder_name(self, folder_name: str):
        self.record_method('Folder', folder_name)

    def set_type(self, wire_type: str):
        self.record_method('Type', wire_type)

    def set_bond_wire_type(self, bond_wire_type: str):
        self.record_method('BondWireType', bond_wire_type)

    def set_height(self, height: float):
        self.record_method('Height', height)

    def set_relative_center_position(self, pos: float):
        self.record_method('RelativeCenterPosition', pos)

    def set_point1(self, x: float, y: float, z: float):
        self.record_method('Point1', x, y, z, False)

    def set_pick_as_point1(self):
        self.record_method('Point1', 0, 0, 0, True)

    def set_point2(self, x: float, y: float, z: float):
        self.record_method('Point2', x, y, z, False)

    def set_pick_as_point2(self):
        self.record_method('Point2', 0, 0, 0, True)

    def set_alpha_angle_deg(self, angle_deg: float):
        self.record_method('Alpha', angle_deg)

    def set_alpha_angle_rad(self, angle_rad: float):
        self.set_alpha_angle_deg(np.rad2deg(angle_rad))

    def set_beta_angle_deg(self, angle_deg: float):
        self.record_method('Beta', angle_deg)

    def set_beta_angle_rad(self, angle_rad: float):
        self.set_beta_angle_deg(np.rad2deg(angle_rad))

    def set_curve_name(self, name: str):
        self.record_method('Curve', name)

    def set_radius(self, radius: float):
        self.record_method('Radius', radius)

    def set_solid_wire_model(self, flag: bool = True):
        self.record_method('SolidWireModel', flag)

    def set_material(self, mat_name: str):
        self.record_method('Material', mat_name)

    def change_material(self, wire_name: str, material_name: str):
        self.record_method('ChangeMaterial', wire_name, material_name)

    def set_termination(self, termination: str):
        self.record_method('Termination', termination)

    def enable_advanced_chain_selection(self, flag: bool = True):
        self.record_method('AdvancedChainSelection', flag)

    def create(self):
        self.record_method('Add')

    def set_solid_name(self, name: str):
        self.record_method('SolidName', name)

    def keep_wire_after_conversion(self, flag: bool = True):
        self.record_method('KeepWire', flag)

    def convert_to_solid_shape(self):
        self.record_method('ConvertToSolidShape')

    def delete(self, wire_name: str):
        self.record_method('Delete', wire_name)

    def rename(self, old_name: str, new_name: str):
        self.record_method('Rename', old_name, new_name)

    def set_automesh_fixpoints(self, wire_name: str, flag: bool = True):
        self.record_method('SetAutomeshFixpoints', wire_name, flag)

    def set_priority(self, wire_name: str, priority: float):
        self.record_method('SetPriority', wire_name, priority)

    def set_material_based_refinement(self, wire_name: str, flag: bool = True):
        self.record_method('SetMaterialBasedRefinement', wire_name, flag)

    def set_mesh_step_width(self, wire_name: str, x: float, y: float, z: float):
        self.record_method('SetMeshStepwidth', wire_name, x, y, z)

    def set_mesh_step_width_tet(self, wire_name: str, step_width: float):
        #TODO: find out the correct name (instead of abbrev)
        self.record_method('SetMeshStepwidthTet', wire_name, step_width)

    def set_mesh_step_width_surface(self, wire_name: str, step_width: float):
        self.record_method('SetMeshStepwidthSrf', wire_name, step_width)

    def set_mesh_extend_with(self, x: float, y: float, z: float):
        self.record_method('SetMeshExtendwidth', x, y, z)

    def activate_mesh_refinement(self, factor: float):
        self.record_method('SetMeshRefinement', True, factor)

    def deactivate_mesh_refinement(self):
        self.record_method('SetMeshRefinement', False, 1.0)

    def activate_mesh_volume_refinement(self, factor: float):
        self.record_method('SetMeshVolumeRefinement', True, factor)

    def deactivate_mesh_volume_refinement(self):
        self.record_method('SetMeshVolumeRefinement', False, 1.0)

    def set_use_for_simulation(self, flag: bool = True):
        self.record_method('SetUseForSimulation', flag)

    def get_length(self, wire_name: str) -> float:
        return self.query_method_float('GetLength', wire_name)

    def get_grid_length(self, wire_name: str) -> float:
        return self.query_method_float('GetGridLength', wire_name)

    def create_folder(self, name: str):
        self.record_method('NewFolder', name)

    def delete_folder(self, name: str):
        self.record_method('DeleteFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.record_method('RenameFolder', old_name, new_name)

    def does_folder_exist(self, name: str) -> bool:
        return self.query_method_bool('DoesFolderExist', name)

    def slice(self, name: str):
        self.record_method('Slice', name)

    def slice_folder(self, name: str):
        self.record_method('SliceFolder', name)


import matplotlib.pyplot as plt
import matplotlib.colors
import numpy as np

def plot_farfield_lin(E_mag, theta_vec_deg, phi_vec_deg, units=None, colormap='turbo'):
    r = E_mag - E_mag.min()
    ax = plt.axes(projection='3d')
    _draw_circles(ax)
    _draw_x_y_z_labels(ax)
    _draw_3d_plot(E_mag, theta_vec_deg, phi_vec_deg, r, ax, units, colormap)

def plot_farfield_log(ff, theta_vec_deg, phi_vec_deg, dyn_range, units=None, colormap='turbo'):
    ff = np.abs(ff)
    ff = _get_field_in_dB(ff, dyn_range)
    plot_farfield_lin(ff, theta_vec_deg, phi_vec_deg, units=units, colormap=colormap)

def _sph2cart(theta_rad, phi_rad, r):
    x = r * np.sin(theta_rad) * np.cos(phi_rad)
    y = r * np.sin(theta_rad) * np.sin(phi_rad)
    z = r * np.cos(theta_rad)
    return x, y, z

def _draw_3d_plot(E_mag, theta_vec_deg, phi_vec_deg, r, ax: plt.Axes, units=None, colormap='turbo'):
    vals_phi, vals_theta = np.meshgrid(phi_vec_deg, theta_vec_deg)
    theta_rad = np.deg2rad(vals_theta)
    phi_rad = np.deg2rad(vals_phi)
    x, y, z = _sph2cart(theta_rad, phi_rad, r/r.max())
    surf = ax.plot_surface(
        x, y, z, cmap=colormap, edgecolor='none', rstride=1, cstride=1, alpha=None,
        antialiased=True, zorder=0.5, linewidth=0)
    ax.set_aspect('equal')
    ax.axis('off')
    ax.grid(visible=False)
    ax.get_figure().tight_layout()
    zoom = 1.2
    ax.get_figure().subplots_adjust(
        left=-1.1-(zoom-1), right=1.0+(zoom-1), top=1.0+(zoom-1), bottom=-0.05-(zoom-1))
    cmap = plt.get_cmap(colormap, 10)
    norm = matplotlib.colors.Normalize(vmin=E_mag.min(), vmax=E_mag.max())
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cb = ax.get_figure().colorbar(
        sm, ticks=np.linspace(E_mag.min(), E_mag.max(), 10), shrink=0.55, pad=-0.05)
    if units is not None:
        cb.ax.set_title(units, y=1.02)
    return surf

def _draw_circles(ax: plt.Axes, radius=1.05):
    angles = np.linspace(0, 2*np.pi, 360)
    a = radius*np.cos(angles)
    b = radius*np.sin(angles)
    c = np.zeros(len(angles))
    ax.plot(a, b, c, linewidth=2, color='b', zorder=0) # z
    ax.plot(c, a, b, linewidth=2, color='r', zorder=0) # x
    ax.plot(b, c, a, linewidth=2, color='g', zorder=0) # y

def _draw_x_y_z_labels(ax: plt.Axes, radius=1.2):
    ax.plot([0, radius], [0, 0], [0, 0], color='r', linewidth=1.5, zorder=0)
    ax.plot([0, 0], [0, radius], [0, 0], color='g', linewidth=1.5, zorder=0)
    ax.plot([0, 0], [0, 0], [0, radius], color='b', linewidth=1.5, zorder=0)
    ax.text(1.1*radius, 0, 0, s='x', zdir=None, zorder=0, weight='bold', color='k')
    ax.text(0, 1.05*radius, 0, s='y', zdir=None, zorder=0, weight='bold', color='k')
    ax.text(0, 0, 1.05*radius, s='z', zdir=None, zorder=0, weight='bold', color='k')

def _get_field_in_dB(f, dyn_range):
    f[f < 0.0000001] = 0.0000001
    f_dB = 20*np.log10(f)
    threshold = f_dB.max()-dyn_range
    f_dB[f_dB < threshold] = threshold
    return f_dB


from . import Project
from . import ComObjectWrapper

class Group(ComObjectWrapper):
    TYPE_NORMAL = 'normal'
    TYPE_MESH = 'mesh'

    def __init__(self, project: Project) -> None:
        self.project = project
        self.com_object = project.com_object.Group

    def invoke_method(self, name, *args, **kwargs):
        self.project.ensure_active()
        return super().invoke_method(name, *args, **kwargs)

    def create(self, name: str, group_type: str = TYPE_NORMAL):
        #BUG: group is not visible after creating (CST Studio Suite bug)
        self.invoke_method('Add', name, group_type)

    def delete(self, name: str):
        self.invoke_method('Delete', name)

    def rename(self, old_name: str, new_name: str):
        self.invoke_method('Rename', old_name, new_name)

    def add_item(self, item_name: str, group_name: str):
        self.invoke_method('AddItem', item_name, group_name)

    def remove_item(self, item_name: str, group_name: str):
        self.invoke_method('RemoveItem', item_name, group_name)

    def create_folder(self, name: str):
        self.invoke_method('NewFolder', name)

    def rename_folder(self, old_name: str, new_name: str):
        self.invoke_method('RenameFolder', old_name, new_name)

    def delete_folder(self, name: str):
        self.invoke_method('DeleteFolder', name)

    def reset(self):
        self.invoke_method('Reset')